import logging
import os

try:
    from pyngrok import ngrok
except ImportError:
    raise ImportError(
        "Please install pyngrok or dbtunnel[ngrok]."
    )

def unsupported_warning(arg):
    if arg is not None:
        print("WARNING: This feature is not supported in ngrok tunnel yet")
    return arg

class NgrokTunnel:

    def __init__(self,

                 port: int,
                 ngrok_tunnel_auth_token: str,
                 ngrok_api_token: str,
                 logger: logging.Logger,
                 basic_auth: str = None,  # "databricks:password"
                 domain: str = None,
                 oauth_provider: str = None,
                 oauth_allow_domains: list[str] = None
                 ):
        self._oauth_allow_domains = unsupported_warning(oauth_allow_domains)
        self._oauth_provider = unsupported_warning(oauth_provider)
        self._domain = unsupported_warning(domain)
        self._basic_auth = unsupported_warning(basic_auth)
        self._ngrok_api_token = ngrok_api_token
        self._ngrok_auth_token = ngrok_tunnel_auth_token
        self._port = port
        self._env = os.environ.copy()
        self._log: logging.Logger = logger

    def _install(self):
        ngrok.install_ngrok()
        ngrok.set_auth_token(self._ngrok_auth_token)

    def run(self):
        self._install()
        listener = ngrok.connect(str(self._port))
        self._log.info(f"Use this information to publicly access your app: \n{listener.public_url}")
        self._log.info("\n\n\n\n")
        return

    def kill_existing_sessions(self):
        # pyngrok doesnt do this so we have to manually do this
        import ngrok
        import requests
        client = ngrok.Client(self._ngrok_api_token)
        for session in client.tunnel_sessions.list():
            session_id = session.id
            self._log.info(f"Killing session: {session_id}")
            stop_session_url = f"https://api.ngrok.com/tunnel_sessions/{session_id}/stop"
            requests.post(stop_session_url, data="{}", headers={
                "Authorization": f"Bearer {self._ngrok_api_token}",
                "Ngrok-Version": "2",
                "Content-Type": "application/json"
            })



import functools
import re

from dbtunnel.vendor.asgiproxy.config import BaseURLProxyConfigMixin, ProxyConfig


def _make_chainlit_local_proxy_config(
        url_base_path: str,
        service_host: str = "0.0.0.0",
        service_port: int = 9989,
        auth_config: dict = None
):
    from dbtunnel.vendor.asgiproxy.config import BaseURLProxyConfigMixin, ProxyConfig
    proxy_root_path = url_base_path
    auth_config = auth_config or {}

    def _modify_root(content, root_path):
        list_of_uris = [b"/assets", b"/public", b"/favicon"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        return content

    def _modify_js_content_root_rewrite(content):
        regex_pattern = r'\{path:"\/",element:(\w+)\.jsx\((\w+),\{\}\)\}'

        decoded_content = content.decode("utf-8")
        # Find all matches
        # find the default root function
        matches = re.findall(regex_pattern, decoded_content)
        if matches:
            jsx_call = matches[0][0]
            func = matches[0][1]  # Assuming there is only one match
            modified_code = re.sub(r'\{path:"\*",element:.*\.jsx\(.*,\{replace:!0,to:"\/"\}\)\}',
                                   f'{{path:"*",element:{jsx_call}.jsx({func},{{}})}}', decoded_content)
            return modified_code.encode("utf-8")
        else:
            print("No match found.")
            return content

    def _modify_js_bundle(content, root_path):
        list_of_uris = [b"/project/settings", b"/auth/config", b"/ws/socket.io", b"/logo", b"/readme"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)

        content = content.replace(b'to:"/",', f'to:"{root_path}",'.encode("utf-8"))
        content = _modify_js_content_root_rewrite(content)
        return content

    def modify_css_bundle(content):
        add_to_end = b" #new-chat-button {display: none;}"
        return content + add_to_end

    def _modify_settings(content, root_path):
        list_of_uris = [b"/public"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        return content

    modify_root = functools.partial(_modify_root, root_path=proxy_root_path)
    modify_js_bundle = functools.partial(_modify_js_bundle, root_path=proxy_root_path)
    modify_settings = functools.partial(_modify_settings, root_path=proxy_root_path)

    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": f"http://{service_host}:{service_port}",
            "rewrite_host_header": f"{service_host}:{service_port}",
            "modify_content": {
                "/": modify_root,
                "": modify_root,
                "*assets/index-*.js": modify_js_bundle,
                "*settings": modify_settings,
                "*assets/index-*.css": modify_css_bundle,
            },
            **auth_config
        },
    )()
    return config


def _make_streamlit_local_proxy_config(
        url_base_path: str,  # noqa
        service_host: str = "0.0.0.0",
        service_port: int = 9989,
        auth_config: dict = None
):
    auth_config = auth_config or {}

    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": f"http://{service_host}:{service_port}",
            "rewrite_host_header": f"{service_host}:{service_port}",
            **auth_config,
        },
    )()
    return config


def _make_gradio_local_proxy_config(
        url_base_path,
        service_host: str = "0.0.0.0",
        service_port: int = 9989,
        auth_config: dict = None
):
    auth_config = auth_config or {}

    def _modify_js_bundle(content, root_path):
        list_of_uris = [b"/theme.css", b"/info", b"/queue", b"/assets"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.rstrip("/").encode("utf-8") + uri)

        content = content.replace(b'to:"/",', f'to:"{root_path}",'.encode("utf-8"))
        return content

    modify_js_bundle = functools.partial(_modify_js_bundle, root_path=url_base_path)

    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": f"http://{service_host}:{service_port}",
            "rewrite_host_header": f"{service_host}:{service_port}",
            "modify_content": {
                "*assets/index-*.js": modify_js_bundle,
                # some reason gradio also has caps index bundled calling out explicitly
                "*assets/Index-*.js": modify_js_bundle,
            },
            **auth_config,
        },
    )()
    return config


def _make_arize_phoenix_local_proxy_config(
        url_base_path,
        service_host: str = "0.0.0.0",
        service_port: int = 9989,
        auth_config: dict = None
):
    auth_config = auth_config or {}

    def _modify_root(content, root_path):
        list_of_uris = [b"/index.css", b"/modernizr.js", b"/favicon.ico", b"/index.js", b"/graphql", b"/projects"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        return content

    def _modify_js_bundle(content, root_path):
        list_of_uris = [b"/graphql", b"/projects"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.rstrip("/").encode("utf-8") + uri)

        return content

    modify_root = functools.partial(_modify_root, root_path=url_base_path)
    modify_js_bundle = functools.partial(_modify_js_bundle, root_path=url_base_path)

    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": f"http://{service_host}:{service_port}",
            "rewrite_host_header": f"{service_host}:{service_port}",
            "modify_content": {
                "/": modify_root,
                "/projects/": modify_root,
                "/projects/*": modify_root,
                "*/index.js": modify_js_bundle,
                # some reason gradio also has caps index bundled calling out explicitly
            },
            **auth_config,
        },
    )()
    return config

class Frameworks:
    STREAMLIT: str = "streamlit"
    GRADIO: str = "gradio"
    CHAINLIT: str = "chainlit"
    ARIZE_PHOENIX: str = "arize_phoenix"


framework_specific_proxy_config = {
    Frameworks.STREAMLIT: _make_streamlit_local_proxy_config,
    Frameworks.GRADIO: _make_gradio_local_proxy_config,
    Frameworks.CHAINLIT: _make_chainlit_local_proxy_config,
    Frameworks.ARIZE_PHOENIX: _make_arize_phoenix_local_proxy_config,
}



# Databricks notebook source
# MAGIC %pip install dbtunnel[nicegui]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from nicegui import ui
from nicegui.events import ValueChangeEventArguments

def show(event: ValueChangeEventArguments):
    name = type(event.sender).__name__
    ui.notify(f'{name}: {event.value}')

ui.button('Button', on_click=lambda: ui.notify('Click'))
with ui.row():
    ui.checkbox('Checkbox', on_change=show)
    ui.switch('Switch', on_change=show)
ui.radio(['A', 'B', 'C'], value='A', on_change=show).props('inline')
with ui.row():
    ui.input('Text input', on_change=show)
    ui.select(['One', 'Two'], value='One', on_change=show)
ui.link('And many more...', '/documentation').classes('mt-8')

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.nicegui(ui).run()

# COMMAND ----------





import tempfile

from dbtunnel.tunnels import DbTunnel
from dbtunnel.utils import execute


class StableDiffusionUITunnel(DbTunnel):

    def __init__(self, no_gpu: bool, port: int, enable_insecure_extensions: bool, extra_flags: str):
        super().__init__(port, "stable-diffusion-ui")
        self._no_gpu = no_gpu
        self._enable_insecure_extensions = enable_insecure_extensions
        self._extra_flags = extra_flags

    def _imports(self):
        return None

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        import requests
        import os
        script_url = "https://raw.githubusercontent.com/AUTOMATIC1111/stable-diffusion-webui/master/webui.sh"

        # Download the script
        response = requests.get(script_url)
        script_content = response.text

        # Create a temporary file to save the script
        with tempfile.NamedTemporaryFile(mode='w+', delete=False) as script_file:
            script_file.write(script_content)
            script_path = script_file.name

        # Set environment variable
        os.environ["COMMANDLINE_ARGS"] = ""

        # skip adding base path if sharing out via tunneling tech
        if self.shared is False:
            os.environ["COMMANDLINE_ARGS"] += f"--subpath={self._proxy_settings.url_base_path.rstrip('/').lstrip('/')} "

        # Set environment variable
        if self._no_gpu:
            os.environ["COMMANDLINE_ARGS"] += (
                f"--skip-torch-cuda-test ")
        # else:
        #     os.environ["COMMANDLINE_ARGS"] = f"--subpath={self._proxy_settings.url_base_path.rstrip('/').lstrip('/')}"

        if self._enable_insecure_extensions:
            os.environ["COMMANDLINE_ARGS"] += " --enable-insecure-extension-access "

        if len(self._extra_flags) > 0:
            os.environ["COMMANDLINE_ARGS"] += f" {self._extra_flags}"

        import os
        import subprocess
        my_env = os.environ.copy()
        subprocess.run(f"kill -9 $(lsof -t -i:{self._port})", capture_output=True, shell=True)

        if self.shared is False:
            self._log.info(f"Deploying stable diffusion web ui app at path: \n{self._proxy_settings.proxy_url}")

        cmd = ["bash", script_path, "-f", "--listen"]
        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path)



import os

from dbtunnel.tunnels import DbTunnel, DbTunnelProxy
from dbtunnel.utils import execute
from dbtunnel.vendor.asgiproxy.frameworks import Frameworks


class GradioAppTunnel(DbTunnel):

    def __init__(self, gradio_app,
                 app_path: str,
                 cwd: str = None,
                 port: int = 8080):
        super().__init__(port, flavor="gradio")
        self._app_path = app_path
        self._cwd = cwd
        self._gradio_app = gradio_app

    def _imports(self):
        try:
            from fastapi import FastAPI
            import uvicorn
            import gradio as gr
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have fastapi, nest_asyncio uvicorn, gradio installed. \n"
                           "pip install fastapi nest_asyncio uvicorn gradio")
            raise e

    def _run(self):
        if self._gradio_app is not None:
            self._run_app()
            return

        self._log.info("Starting server...")

        gradio_service_port = 9908
        port = self._port
        url_base_path = self._proxy_settings.url_base_path

        proxy_service = DbTunnelProxy(
            proxy_port=port,
            service_port=gradio_service_port,
            url_base_path=url_base_path,
            framework=Frameworks.GRADIO,
            token_auth=self._basic_tunnel_auth["token_auth"],
            token_auth_workspace_url=self._basic_tunnel_auth["token_auth_workspace_url"],
            cwd=self._cwd
        )

        proxy_service.start()

        self._log.info(f"Use this link to access the Gradio UI in Databricks: \n{self._proxy_settings.proxy_url}")

        self._log.info("Starting gradio...")

        my_env = os.environ.copy()
        my_env["GRADIO_SERVER_PORT"] = str(gradio_service_port)
        my_env["GRADIO_SERVER_NAME"] = "0.0.0.0"

        cmd = ["python", self._app_path]

        self._log.info(f"Running command: {' '.join(cmd)}")
        for log_stmt in execute(cmd, my_env, cwd=self._cwd):
            self._log.info(log_stmt.rstrip("/n"))

        proxy_service.wait()

    def _run_app(self):
        self.display()
        self._log.info("Starting server...")
        from fastapi import FastAPI
        import uvicorn
        import gradio as gr
        app = FastAPI(root_path=self._proxy_settings.url_base_path.rstrip("/"))
        app = gr.mount_gradio_app(app, self._gradio_app, path="/")
        import nest_asyncio
        nest_asyncio.apply()

        # Start the server
        async def start():
            config = uvicorn.Config(
                app,
                host="0.0.0.0",
                port=self._port,
            )
            server = uvicorn.Server(config)
            await server.serve()

        # Run the asyncio event loop instead of uvloop to enable re entrance
        import asyncio
        self._log.info(f"Use this link: \n{self._proxy_settings.proxy_url}")
        asyncio.run(start())

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'



import chainlit as cl


@cl.on_message
async def main(message: cl.Message):
    # Your custom logic goes here...

    # Send a response back to the user
    await cl.Message(
        content=f"Received: {message.content}",
    ).send()



import streamlit as st
import pandas as pd
import numpy as np

st.title('Uber pickups in NYC')

DATE_COLUMN = 'date/time'
DATA_URL = ('https://s3-us-west-2.amazonaws.com/'
            'streamlit-demo-data/uber-raw-data-sep14.csv.gz')

@st.cache
def load_data(nrows):
    data = pd.read_csv(DATA_URL, nrows=nrows)
    lowercase = lambda x: str(x).lower()
    data.rename(lowercase, axis='columns', inplace=True)
    data[DATE_COLUMN] = pd.to_datetime(data[DATE_COLUMN])
    return data

data_load_state = st.text('Loading data...')
data = load_data(10000)
data_load_state.text("Done! (using st.cache)")

if st.checkbox('Show raw data'):
    st.subheader('Raw data')
    st.write(data)

st.subheader('Number of pickups by hour')
hist_values = np.histogram(data[DATE_COLUMN].dt.hour, bins=24, range=(0,24))[0]
st.bar_chart(hist_values)

# Some number in the range 0-23
hour_to_filter = st.slider('hour', 0, 23, 17)
filtered_data = data[data[DATE_COLUMN].dt.hour == hour_to_filter]

st.subheader('Map of all pickups at %s:00' % hour_to_filter)
st.map(filtered_data)


import asyncio
from typing import Optional

import aiohttp

from dbtunnel.vendor.asgiproxy.config import ProxyConfig


class ProxyContext:
    semaphore: asyncio.Semaphore
    _session: Optional[aiohttp.ClientSession] = None

    def __init__(
        self,
        config: ProxyConfig,
        max_concurrency: int = 20,
    ) -> None:
        self.config = config
        self.semaphore = asyncio.Semaphore(max_concurrency)

    @property
    def session(self) -> aiohttp.ClientSession:
        if not self._session:
            self._session = aiohttp.ClientSession(
                cookie_jar=aiohttp.DummyCookieJar(), auto_decompress=False
            )
        return self._session

    async def __aenter__(self) -> "ProxyContext":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # noqa: ANN001
        await self.close()

    async def close(self) -> None:
        if self._session:
            await self._session.close()



from dbtunnel.tunnels import DbTunnel, DbTunnelProxy
from dbtunnel.utils import process_file, execute
from dbtunnel.vendor.asgiproxy.frameworks import Frameworks

class StreamlitTunnel(DbTunnel):

    def __init__(self, script_path: str, port: int):
        super().__init__(port, "streamlit")
        self._script_path = script_path

    def _imports(self):
        try:
            import streamlit
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have nest_asyncio, streamlit installed. \n"
                           "pip install nest_asyncio streamlit")
            raise e

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        import nest_asyncio
        nest_asyncio.apply()

        streamlit_service_port = 9908
        port = self._port
        url_base_path = self._proxy_settings.url_base_path

        proxy_service = DbTunnelProxy(
            proxy_port=port,
            service_port=streamlit_service_port,
            url_base_path=url_base_path,
            framework=Frameworks.STREAMLIT,
            token_auth=self._basic_tunnel_auth["token_auth"],
            token_auth_workspace_url=self._basic_tunnel_auth["token_auth_workspace_url"],
            cwd=None
        )

        proxy_service.start()

        self._log.info(f"Use this link to access the Streamlit UI in Databricks: \n{self._proxy_settings.proxy_url}")

        with process_file(self._script_path) as file_path:
            self.run_streamlit(file_path, streamlit_service_port)

        proxy_service.wait()

    def run_streamlit(self, path, port):
        import os
        import subprocess
        my_env = os.environ.copy()
        my_env["STREAMLIT_SERVER_PORT"] = f"{port}"
        my_env["STREAMLIT_SERVER_ADDRESS"] = "0.0.0.0"
        my_env["STREAMLIT_SERVER_HEADLESS"] = "true"
        subprocess.run(f"kill -9 $(lsof -t -i:{port})", capture_output=True, shell=True)

        self._log.info(f"Deploying streamlit app at path: {path} on port: {port}")
        cmd = [
            "streamlit",
            "run",
            path,
            "--browser.gatherUsageStats",
            "false",
            "--server.fileWatcherType",
            "none"
        ]
        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path.rstrip("\n"))


def streamlit_patch_websockets_v2():
    from pathlib import Path
    import subprocess
    import streamlit
    p = Path(streamlit.__file__)
    _dir = (p.parent)
    process = subprocess.run(f"find {_dir} -type f -exec sed -i -e 's/\"stream\"/\"ws\"/g' {{}} \;",
                             capture_output=True, shell=True)
    return process.stdout.decode()



from typing import Optional, Iterator

from starlette.types import Scope

from dbtunnel.vendor.asgiproxy.frameworks import Frameworks

IS_DATABRICKS_PROXY_SCOPE_KEY = "__is_databricks_proxy"


def get_hosts_from_headers(scope: Scope) -> Iterator[str]:
    for header in scope["headers"]:
        key = header[0].decode("utf-8")
        if key.lower() == "host":
            yield header[1].decode("utf-8")
        if key.lower() == "x-forwarded-host":
            yield header[1].decode("utf-8")
    return

def get_forwarded_host_from_headers(scope: Scope) -> Optional[str]:
    for header in scope["headers"]:
        key = header[0].decode("utf-8")
        if key.lower() == "x-forwarded-host":
            return header[1].decode("utf-8")
    return None


def is_databricks_host(host: str) -> bool:
    # local host is for testing purposes
    return host.endswith(".azuredatabricks.net") or host.endswith(".databricks.com") \
        or host.startswith("0.0.0.0") or host.startswith("127.0.0.1") or host.startswith("localhost")


def add_if_databricks_proxy_scope(scope: Scope) -> None:
    if IS_DATABRICKS_PROXY_SCOPE_KEY in scope:
        return
    hosts = get_hosts_from_headers(scope)
    if any(is_databricks_host(host) for host in hosts) is True:
        scope[IS_DATABRICKS_PROXY_SCOPE_KEY] = True


def is_from_databricks_proxy(scope: Scope) -> bool:
    return IS_DATABRICKS_PROXY_SCOPE_KEY in scope and scope[IS_DATABRICKS_PROXY_SCOPE_KEY] is True


FRAMEWORK_KEY = "__framework"
PROXY_PORT = "__proxy_port"

def add_framework_to_scope(scope: Scope, framework: str) -> None:
    scope[FRAMEWORK_KEY] = framework
    return

def add_origin_port_to_scope(scope: Scope, port: int) -> None:
    scope[PROXY_PORT] = str(port)
    return


def get_origin_port_from_scope(scope: Scope):
    return scope.get(PROXY_PORT)

def is_streamlit(scope: Scope) -> bool:
    return Frameworks.STREAMLIT == scope.get(FRAMEWORK_KEY, None)


# Databricks notebook source
# MAGIC %pip install dbtunnel[gradio]

# COMMAND ----------

# MAGIC %pip install -U databricks-sdk databricks-sql-connector

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

# MAGIC %md
# MAGIC
# MAGIC ## Get the first serverless warehouse

# COMMAND ----------

from dbtunnel import ctx, compute_utils

# defaults to serverless only and ignores case
# just run the first warehouse
details = compute_utils.get_warehouse("*", ignore_case=True, serverless_only=True)
if details is None:
  raise Exception("Unable to find a cluster with the following setting")
hostname, path = details.hostname, details.http_path
# printing the name dont really need it anywhere else
hostname, path, details.name

# COMMAND ----------

# MAGIC %md
# MAGIC
# MAGIC ## Setup

# COMMAND ----------

from databricks import sql
import os

CATALOG = "main"
SCHEMA = "default"

with sql.connect(server_hostname = hostname,
                 http_path       = path,
                 access_token    = ctx.token) as connection:
  with connection.cursor() as cursor:
    cursor.execute(f"""CREATE OR REPLACE TABLE {CATALOG}.{SCHEMA}.trips_example AS
SELECT * FROM samples.nyctaxi.trips LIMIT 100;""")
  with connection.cursor() as cursor:
    cursor.execute(f"""CREATE TABLE IF NOT EXISTS {CATALOG}.{SCHEMA}.trips_feedback (
  email STRING,
  feedback STRING,
  datetime STRING
) USING DELTA;""")


# COMMAND ----------

# MAGIC %md
# MAGIC
# MAGIC ## Handler functions

# COMMAND ----------

import pandas as pd
def get_trips_df(token):
  with sql.connect(server_hostname = hostname,
                 http_path       = path,
                 access_token    = token) as connection:
    with connection.cursor() as cursor:
      cursor.execute(f"SELECT * FROM {CATALOG}.{SCHEMA}.trips_example")
      # Fetch the result
      result = cursor.fetchall()

      # Convert the result to a DataFrame
      columns = [column[0] for column in cursor.description]
      return pd.DataFrame(result, columns=columns)

def get_feedback_df(token):
  with sql.connect(server_hostname = hostname,
                 http_path       = path,
                 access_token    = token) as connection:
    with connection.cursor() as cursor:
      cursor.execute(f"SELECT * FROM {CATALOG}.{SCHEMA}.trips_feedback")
      # Fetch the result
      result = cursor.fetchall()

      # Convert the result to a DataFrame
      columns = [column[0] for column in cursor.description]
      return pd.DataFrame(result, columns=columns)
    
from datetime import datetime
def insert_feedback(email: str, feedback: str, token: str):
  with sql.connect(server_hostname = hostname,
                 http_path       = path,
                 access_token    = token) as connection:
    with connection.cursor() as cursor:
      cursor.execute(f"INSERT INTO {CATALOG}.{SCHEMA}.trips_feedback VALUES (?, ?, ?)", (email, feedback, datetime.now().strftime("%Y-%m-%d %H:%M:%S")))


# COMMAND ----------

import gradio as gr
import pandas as pd
from databricks.sdk import WorkspaceClient
import os

def get_current_user(token: str):
  if token is None:
    return None
  try:
      # based on user's input token
      return WorkspaceClient(host=ctx.host,token=token).current_user.me().user_name
  except Exception as e:
      print(e)
      return None

def submit(token: str):
    user = get_current_user(token)
    if user is None:
      return None, None, "Invalid Token"
    return get_trips_df(token), get_feedback_df(token), user


def insert_review(feedback: str, token: str):
  insert_feedback(get_current_user(token), feedback, token)
  return get_feedback_df(token)


with gr.Blocks() as demo:

    token_field = gr.Textbox(label="Token", type="password", lines=1)
    workspace = gr.Textbox(label="Workspace", value=ctx.host, lines=1)
    logged_in_as = gr.Textbox(label="Email", lines=1)
    with gr.Row():
      data = gr.DataFrame(label="Raw Data")
      feedback_display = gr.DataFrame(label="Feedback Data")
    btn = gr.Button(value="Submit")
    btn.click(submit, inputs=[token_field], outputs=[data, feedback_display, logged_in_as])

    with gr.Row():
        feedback = gr.Textbox(label="Feedback")
    
    feedback_btn = gr.Button(value="Submit Feedback")
    feedback_btn.click(insert_review, inputs=[feedback, token_field], outputs=[feedback_display])


# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.gradio(demo).run()

# COMMAND ----------





# Databricks notebook source
# MAGIC %pip install dbtunnel[dash]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from dash import Dash, html, dcc, callback, Output, Input
import plotly.express as px
import pandas as pd

df = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/gapminder_unfiltered.csv')

app = Dash("app")

app.layout = html.Div([
    html.H1(children='Title of Dash App', style={'textAlign':'center'}),
    dcc.Dropdown(df.country.unique(), 'Canada', id='dropdown-selection'),
    dcc.Graph(id='graph-content')
])

@callback(
    Output('graph-content', 'figure'),
    Input('dropdown-selection', 'value')
)
def update_graph(value):
    dff = df[df.country==value]
    return px.line(dff, x='year', y='pop')

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.dash(app).run()

# COMMAND ----------





# Databricks notebook source
# MAGIC %pip install dbtunnel[shiny]

# COMMAND ----------

import matplotlib.pyplot as plt
import numpy as np
from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.panel_sidebar(
            ui.input_slider("n", "N", 0, 100, 20),
        ),
        ui.panel_main(
            ui.output_plot("histogram"),
        ),
    ),
)


def server(input, output, session):
    @output
    @render.plot(alt="A histogram")
    def histogram():
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(437)
        plt.hist(x, input.n(), density=True)


app = App(app_ui, server, debug=True)

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.shiny_python(app, port=7657).run()



import gradio as gr
import os


def combine(a, b):
    return a + " " + b


def mirror(x):
    return x


with gr.Blocks() as demo:

    txt = gr.Textbox(label="Input", lines=2)
    txt_2 = gr.Textbox(label="Input 2")
    txt_3 = gr.Textbox(value="", label="Output")
    btn = gr.Button(value="Submit")
    btn.click(combine, inputs=[txt, txt_2], outputs=[txt_3])

    with gr.Row():
        im = gr.Image()
        im_2 = gr.Image()

    btn = gr.Button(value="Mirror Image")
    btn.click(mirror, inputs=[im], outputs=[im_2])

    gr.Markdown("## Text Examples")
    gr.Examples(
        [["hi", "Adam"], ["hello", "Eve"]],
        [txt, txt_2],
        txt_3,
        combine,
        cache_examples=True,
    )

if __name__ == "__main__":
  demo.queue().launch(
    server_name=os.getenv("GRADIO_SERVER_NAME"), 
    server_port=int(os.getenv("GRADIO_SERVER_PORT")),
  )


# Databricks notebook source
# MAGIC %pip install dbtunnel[asgiproxy,chainlit]

# COMMAND ----------

# MAGIC %pip install databricks-genai-inference langchain-community tiktoken langchain mlflow-skinny chromadb rank_bm25 simsimd

# COMMAND ----------

# MAGIC %pip install -U langchain==0.1.3 aiohttp sqlalchemy

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

# MAGIC %md
# MAGIC
# MAGIC ### RAG Demo using Foundation Model APIs (Mixtral 8x7b Instruct Chat + BGE-en-Large Embeddings)
# MAGIC
# MAGIC Take a look at the langchain code in the file `chainlit_foundation_model.py`.

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/chainlit_foundation_model.py"

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.chainlit(script_path).inject_auth().run()

# COMMAND ----------





import asyncio
import logging
import uuid
from typing import Optional

from aiohttp import ClientWebSocketResponse, WSMessage, WSMsgType
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket
from websockets.exceptions import ConnectionClosed

from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.utils.headers import is_from_databricks_proxy, is_streamlit, get_origin_port_from_scope

log = logging.getLogger(__name__)


class UnknownMessage(ValueError):
    pass


class WebSocketProxyContext:
    def __init__(
            self,
            *,
            id: str = None,
            client_ws: WebSocket,
            upstream_ws: ClientWebSocketResponse,
    ) -> None:
        self.id = str(id or uuid.uuid4())
        self.client_ws = client_ws
        self.upstream_ws = upstream_ws

    async def client_to_upstream_loop(self):
        while True:
            client_msg: dict = await self.client_ws.receive()
            if client_msg["type"] == "websocket.disconnect":
                log.info(f"WSP {self.id}: Client closed connection.")
                return
            log.debug(f"C->U: {client_msg}")
            await self.send_client_to_upstream(client_msg)

    async def send_client_to_upstream(self, client_msg: dict):
        if client_msg.get("text"):
            await self.upstream_ws.send_str(client_msg["text"])
            return True

        if client_msg.get("bytes"):
            await self.upstream_ws.send_bytes(client_msg["bytes"])
            return True

        raise UnknownMessage(
            f"WSP {self.id}: Unknown client WS message: {client_msg}", client_msg
        )

    async def send_upstream_to_client(self, upstream_msg: WSMessage):
        if upstream_msg.type == WSMsgType.text:
            await self.client_ws.send_text(upstream_msg.data)
            return
        if upstream_msg.type == WSMsgType.binary:
            await self.client_ws.send_bytes(upstream_msg.data)
            return
        raise UnknownMessage(
            f"WSP {self.id}: Unknown upstream WS message: {upstream_msg}", upstream_msg
        )

    async def upstream_to_client_loop(self):
        while True:
            upstream_msg: WSMessage = await self.upstream_ws.receive()
            log.debug(f"WSP {self.id}: U->C: {upstream_msg}")

            if upstream_msg.type == WSMsgType.closed:
                log.info(f"WSP {self.id}: Upstream closed connection.")
                return

            try:
                await self.send_upstream_to_client(upstream_msg=upstream_msg)
            except ConnectionClosed as cc:
                log.info(
                    f"WSP {self.id}: Upstream-to-client loop: client connection had closed ({cc})."
                )
                return

    async def loop(self):
        log.debug(f"WSP {self.id}: Starting main loop.")
        ctu_task = asyncio.create_task(self.client_to_upstream_loop())
        utc_task = asyncio.create_task(self.upstream_to_client_loop())
        try:
            await asyncio.wait(
                [ctu_task, utc_task], return_when=asyncio.FIRST_COMPLETED
            )
            ctu_task.cancel()
            utc_task.cancel()
        except Exception:
            log.warning(f"WSP {self.id}: Unexpected exception!", exc_info=True)
            raise
        log.debug(f"WSP {self.id}: Ending main loop.")


async def proxy_websocket(
        *, context: ProxyContext, scope: Scope, receive: Receive, send: Send
) -> None:
    # query params are important for socket.io for websocket upgrade
    if is_from_databricks_proxy(scope) is True:
        root_path = scope["root_path"]
        if scope["path"].startswith(root_path):
            scope["path"] = scope["path"].replace(root_path, "")

    q_string = scope.get("query_string", None)

    # ensure query params it is important for socketio during websocket upgrade
    if q_string is not None and q_string.decode("utf-8") not in scope["path"]:
        scope["path"] = scope["path"] + "?" + q_string.decode("utf-8")

    if is_streamlit(scope) is True:
        def handle_header(header):
            header_decoded = header[0].decode("utf-8").lower()
            # these are important headers for streamlit
            if header_decoded in ["origin"]:
                return header[0], f'http://0.0.0.0:{get_origin_port_from_scope(scope)}'.encode("utf-8")
            if header_decoded in ["accept-encoding"]:
                return header[0], b"gzip, deflate"
            return header

        # remove all x- headers for streamlit and cf- headers
        scope["headers"] = [handle_header(header) for header in scope["headers"] if
                            not header[0].decode("utf-8").startswith("x-") and not header[0].decode("utf-8").startswith(
                                "cf-")]

    client_ws: Optional[WebSocket] = None
    upstream_ws: Optional[ClientWebSocketResponse] = None
    try:
        client_ws = WebSocket(scope=scope, receive=receive, send=send)
        ctx = context.config.get_upstream_websocket_options(
            scope=scope, client_ws=client_ws
        )
        async with context.session.ws_connect(
                **ctx
        ) as upstream_ws:
            await client_ws.accept(subprotocol=upstream_ws.protocol)
            ctx = WebSocketProxyContext(client_ws=client_ws, upstream_ws=upstream_ws)
            await ctx.loop()
    finally:
        if upstream_ws:
            try:
                await upstream_ws.close()
            except Exception:
                pass
        if client_ws:
            try:
                await client_ws.close()
            except Exception:
                pass



# Databricks notebook source
# MAGIC %pip install dbtunnel[flask]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from flask import Flask

app = Flask("test")

@app.route('/')
def hello_world():
    return 'Hello, World!'


# COMMAND ----------

from dbtunnel import dbtunnel

dbtunnel.flask(app).run()

# COMMAND ----------





# Databricks notebook source
# MAGIC %pip install dbtunnel

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.stable_diffusion_ui(no_gpu=True).run()



import asyncio
from typing import Union

import aiohttp

Streamable = Union[asyncio.StreamReader, aiohttp.StreamReader]


async def read_stream_in_chunks(stream: Streamable, chunk_size: int = 524_288):
    while True:
        chunk = await stream.read(chunk_size)
        yield chunk
        if not chunk:
            break



# Databricks notebook source
# MAGIC %pip install dbtunnel[asgiproxy,streamlit,ngrok]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/streamlit_example.py"

# COMMAND ----------

from dbtunnel import dbtunnel

# Go to [ngrok](https://ngrok.com/) and create an account and get an api token
# get ngrok token from secrets or hardcode it
# you get ngrok tunnel auth token from here: https://dashboard.ngrok.com/get-started/your-authtoken
# you get the ngrok api token from here: https://dashboard.ngrok.com/api
# you must have both so that we can kill external sessions if this is a personal account
dbtunnel.streamlit(script_path).share_to_internet_via_ngrok(
    ngrok_api_token="<ngrok api token>",
    ngrok_tunnel_auth_token="<ngrok tunnel auth token>",
    # kill_all_tunnel_sessions=True,
).run()



from dbtunnel.tunnels import DbTunnel
from dbtunnel.utils import execute, process_file


class BokehTunnel(DbTunnel):

    def __init__(self, script_path: str, port: int):
        super().__init__(port, "bokeh")
        self._script_path = script_path

    def _imports(self):
        try:
            import nest_asyncio
            import bokeh
        except ImportError as e:
            self._log.info("ImportError: Make sure you have nest_asyncio, bokeh installed. \n"
                  "pip install nest_asyncio bokeh")
            raise e

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        import nest_asyncio
        nest_asyncio.apply()
        self._log.info(f"Use this link: \n{self._proxy_settings.proxy_url}")
        with process_file(self._script_path) as file_path:
            self.run_bokeh(file_path, self._port)

    def run_bokeh(self, path, port):
        import os
        import subprocess
        my_env = os.environ.copy()
        subprocess.run(f"kill -9 $(lsof -t -i:{port})", capture_output=True, shell=True)
        # static assets otherwise get served by root and the root path is not allowed!
        my_env["BOKEH_RESOURCES"] = "cdn"
        self._log.info(f"Deploying {self._flavor} app at path: {path} on port: {port}")
        server_path_prefix = self._proxy_settings.url_base_path.rstrip('/').lstrip('/')
        cmd = ["bokeh",
               "serve",
               "--show",
               path,
               "--port",
               str(port),
               # "--prefix",
               # server_path_prefix
               ]
        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path)



from typing import Optional, Iterable, Dict, Callable
from urllib.parse import urljoin

import aiohttp
from multidict import MultiDict, CIMultiDict
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.types import Scope
from starlette.websockets import WebSocket

Headerlike = MultiDict


class ProxyConfig:
    def get_upstream_url(self, *, scope: Scope) -> str:
        """
        Get the upstream URL for a client request.
        """
        raise NotImplementedError("...")

    def get_client_protocols(self, *, scope: Scope, headers: Headers) -> Iterable[str]:
        """
        Get client subprotocol list so it can be passed upstream.
        """
        return scope.get("subprotocols", [])

    def get_upstream_url_with_query(self, *, scope: Scope) -> str:
        """
        Get the upstream URL for a client request, including any query parameters to include.
        """
        # The default implementation simply appends the original URL's query string to the
        # upstream URL generated by `get_upstream_url`.
        url = self.get_upstream_url(scope=scope)
        query_string = scope.get("query_string")
        if query_string:
            sep = "&" if "?" in url else "?"
            url += "{}{}".format(sep, query_string.decode("utf-8"))
        return url

    def process_client_headers(self, *, scope: Scope, headers: Headers) -> Headerlike:
        """
        Process client HTTP headers before they're passed upstream.
        """
        return headers

    def process_upstream_headers(
            self, *, scope: Scope, proxy_response: aiohttp.ClientResponse
    ) -> Headerlike:
        """
        Process upstream HTTP headers before they're passed to the client.
        """
        headers = CIMultiDict(proxy_response.headers)
        return headers  # type: ignore

    def get_upstream_http_options(
            self, *, scope: Scope, client_request: Request, data
    ) -> dict:
        """
        Get request options (as passed to aiohttp.ClientSession.request).
        """
        return dict(
            method=client_request.method,
            url=self.get_upstream_url_with_query(scope=scope),
            data=data,
            headers=self.process_client_headers(
                scope=scope,
                headers=client_request.headers,
            ),
            allow_redirects=False,
        )

    def get_upstream_websocket_options(
            self, *, scope: Scope, client_ws: WebSocket
    ) -> dict:
        """
        Get websocket connection options (as passed to aiohttp.ClientSession.ws_connect).
        """
        return dict(
            method=scope.get("method", "GET"),
            url=self.get_upstream_url(scope=scope),
            headers=self.process_client_headers(scope=scope, headers=client_ws.headers),
            protocols=self.get_client_protocols(scope=scope, headers=client_ws.headers),
        )


class BaseURLProxyConfigMixin:
    upstream_base_url: str
    rewrite_host_header: Optional[str] = None
    modify_content: Optional[Dict[str, Callable[[str], str]]] = None
    token_auth: Optional[bool] = False
    token_auth_workspace_url: Optional[str] = None

    def get_upstream_url(self, scope: Scope) -> str:
        return urljoin(self.upstream_base_url, scope["path"])

    def process_client_headers(
            self, *, scope: Scope, headers: Headerlike
    ) -> Headerlike:
        """
        Process client HTTP headers before they're passed upstream.
        """
        if self.rewrite_host_header:
            headers = headers.mutablecopy()  # type: ignore
            headers["host"] = self.rewrite_host_header
        return super().process_client_headers(scope=scope, headers=headers)  # type: ignore



# Databricks notebook source
# MAGIC %md
# MAGIC # Serving Mistral-7B-Instruct via vllm with a cluster driver proxy app
# MAGIC
# MAGIC The [Mistral-7B-Instruct-v0.1](https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.1) Large Language Model (LLM) is a instruct fine-tuned version of the [Mistral-7B-v0.1](https://huggingface.co/mistralai/Mistral-7B-v0.1) generative text model using a variety of publicly available conversation datasets.
# MAGIC [vllm](https://github.com/vllm-project/vllm/tree/main) is an open-source library that makes LLM inference fast with various optimizations.
# MAGIC
# MAGIC Environment for this notebook:
# MAGIC - Runtime: 14.0 GPU ML Runtime
# MAGIC - Instance: `g5.xlarge` on AWS, `Standard_NV36ads_A10_v5` on Azure
# MAGIC

# COMMAND ----------

# MAGIC %pip install dbtunnel[gradio] vllm==0.2.0 transformers==4.34.0 accelerate==0.20.3

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

# MAGIC %md
# MAGIC ## Inference
# MAGIC The example in the model card should also work on Databricks with the same environment.
# MAGIC

# COMMAND ----------

from vllm import LLM

# it is suggested to pin the revision commit hash and not change it for reproducibility because the uploader might change the model afterwards; you can find the commmit history of Mistral-7B-Instruct-v0. in https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.1/commits/main
model = "mistralai/Mistral-7B-Instruct-v0.1"
revision = "3dc28cf29d2edd31a0a7b8f0b21637059815b4d5"

llm = LLM(model=model, revision=revision)

# COMMAND ----------

from vllm import SamplingParams

# Prompt templates as follows could guide the model to follow instructions and respond to the input, and empirically it turns out to make Falcon models produce better responses
DEFAULT_SYSTEM_PROMPT = """\
You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe. Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.

If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information."""

INTRO_BLURB = "Below is an instruction that describes a task. Write a response that appropriately completes the request."
PROMPT_FOR_GENERATION_FORMAT = """
<s>[INST]<<SYS>>
{system_prompt}
<</SYS>>


{instruction}
[/INST]
""".format(
    system_prompt=DEFAULT_SYSTEM_PROMPT,
    instruction="{instruction}"
)

# Define parameters to generate text
def gen_text_for_serving(prompt, **kwargs):
    prompt = PROMPT_FOR_GENERATION_FORMAT.format(instruction=prompt)

    # the default max length is pretty small (20), which would cut the generated output in the middle, so it's necessary to increase the threshold to the complete response
    if "max_tokens" not in kwargs:
        kwargs["max_tokens"] = 512

    sampling_params = SamplingParams(**kwargs)

    outputs = llm.generate(prompt, sampling_params=sampling_params)
    texts = [out.outputs[0].text for out in outputs]

    return texts[0]


# COMMAND ----------

print(gen_text_for_serving("How to master Python in 3 days?"))

# COMMAND ----------

import gradio as gr
import time

def respond(message, history):
    computed_gen = str(gen_text_for_serving(message))
    for i in range(len(computed_gen)):
          time.sleep(0.05)
          yield computed_gen[: i+1]
    return 

demo = gr.ChatInterface(fn=respond, examples=["hello what is the python language?"], title="Mistral Bot").queue()

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.gradio(demo).run()

# COMMAND ----------





import numpy as np

from bokeh.layouts import gridplot
from bokeh.models import BoxSelectTool, LassoSelectTool
from bokeh.plotting import curdoc, figure

# create three normal population samples with different parameters
x1 = np.random.normal(loc=5.0, size=400) * 100
y1 = np.random.normal(loc=10.0, size=400) * 10

x2 = np.random.normal(loc=5.0, size=800) * 50
y2 = np.random.normal(loc=5.0, size=800) * 10

x3 = np.random.normal(loc=55.0, size=200) * 10
y3 = np.random.normal(loc=4.0, size=200) * 10

x = np.concatenate((x1, x2, x3))
y = np.concatenate((y1, y2, y3))

TOOLS="pan,wheel_zoom,box_select,lasso_select,reset"

# create the scatter plot
p = figure(tools=TOOLS, width=600, height=600, min_border=10, min_border_left=50,
           toolbar_location="above", x_axis_location=None, y_axis_location=None,
           title="Linked Histograms")
p.background_fill_color = "#fafafa"
p.select(BoxSelectTool).select_every_mousemove = False
p.select(LassoSelectTool).select_every_mousemove = False

r = p.scatter(x, y, size=3, color="#3A5785", alpha=0.6)

# create the horizontal histogram
hhist, hedges = np.histogram(x, bins=20)
hzeros = np.zeros(len(hedges)-1)
hmax = max(hhist)*1.1

LINE_ARGS = dict(color="#3A5785", line_color=None)

ph = figure(toolbar_location=None, width=p.width, height=200, x_range=p.x_range,
            y_range=(-hmax, hmax), min_border=10, min_border_left=50, y_axis_location="right")
ph.xgrid.grid_line_color = None
ph.yaxis.major_label_orientation = np.pi/4
ph.background_fill_color = "#fafafa"

ph.quad(bottom=0, left=hedges[:-1], right=hedges[1:], top=hhist, color="white", line_color="#3A5785")
hh1 = ph.quad(bottom=0, left=hedges[:-1], right=hedges[1:], top=hzeros, alpha=0.5, **LINE_ARGS)
hh2 = ph.quad(bottom=0, left=hedges[:-1], right=hedges[1:], top=hzeros, alpha=0.1, **LINE_ARGS)

# create the vertical histogram
vhist, vedges = np.histogram(y, bins=20)
vzeros = np.zeros(len(vedges)-1)
vmax = max(vhist)*1.1

pv = figure(toolbar_location=None, width=200, height=p.height, x_range=(-vmax, vmax),
            y_range=p.y_range, min_border=10, y_axis_location="right")
pv.ygrid.grid_line_color = None
pv.xaxis.major_label_orientation = np.pi/4
pv.background_fill_color = "#fafafa"

pv.quad(left=0, bottom=vedges[:-1], top=vedges[1:], right=vhist, color="white", line_color="#3A5785")
vh1 = pv.quad(left=0, bottom=vedges[:-1], top=vedges[1:], right=vzeros, alpha=0.5, **LINE_ARGS)
vh2 = pv.quad(left=0, bottom=vedges[:-1], top=vedges[1:], right=vzeros, alpha=0.1, **LINE_ARGS)

layout = gridplot([[p, pv], [ph, None]], merge_tools=False)

curdoc().add_root(layout)
curdoc().title = "Selection Histogram"

def update(attr, old, new):
    inds = new
    if len(inds) == 0 or len(inds) == len(x):
        hhist1, hhist2 = hzeros, hzeros
        vhist1, vhist2 = vzeros, vzeros
    else:
        neg_inds = np.ones_like(x, dtype=np.bool)
        neg_inds[inds] = False
        hhist1, _ = np.histogram(x[inds], bins=hedges)
        vhist1, _ = np.histogram(y[inds], bins=vedges)
        hhist2, _ = np.histogram(x[neg_inds], bins=hedges)
        vhist2, _ = np.histogram(y[neg_inds], bins=vedges)

    hh1.data_source.data["top"]   =  hhist1
    hh2.data_source.data["top"]   = -hhist2
    vh1.data_source.data["right"] =  vhist1
    vh2.data_source.data["right"] = -vhist2

r.data_source.selected.on_change('indices', update)


# Databricks notebook source
# MAGIC %md
# MAGIC # Serving Databricks chat completion model
# MAGIC
# MAGIC Using [llama-2-70b-chat](https://docs.databricks.com/en/machine-learning/foundation-models/query-foundation-model-apis.html) chat in a gradio chat app.
# MAGIC
# MAGIC We’re excited to announce that Meta AI’s Llama 2 foundation chat models are available in the Databricks Marketplace for you to fine-tune and deploy on private model serving endpoints. The Databricks Marketplace is an open marketplace that enables you to share and exchange data assets, including datasets and notebooks, across clouds, regions, and platforms. Adding to the data assets already offered on Marketplace, this new listing provides instant access to Llama 2's chat-oriented large language models (LLM), from 7 to 70 billion parameters, as well as centralized governance and lineage tracking in the Unity Catalog. Each model is wrapped in MLflow to make it easy for you to use the MLflow Evaluation API in Databricks notebooks as well as to deploy with a single-click on our LLM-optimized GPU model serving endpoints. 
# MAGIC
# MAGIC https://www.databricks.com/blog/llama-2-foundation-models-available-databricks-lakehouse-ai

# COMMAND ----------

# MAGIC %pip install dbtunnel[gradio] databricks-genai-inference

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

# MAGIC %md
# MAGIC ## Inference
# MAGIC The example in the model card should also work on Databricks with the same environment.
# MAGIC

# COMMAND ----------

from databricks_genai_inference import ChatSession

# change the system prompt

def make_chat_session():
    system_message = "You are a helpful assistant."
    max_tokens = 128
    return ChatSession(model="llama-2-70b-chat", system_message=system_message, max_tokens=max_tokens)

chat = make_chat_session()

# COMMAND ----------

import gradio as gr
import time


def respond(message, history):
    # global so we can reset it
    global chat

    if history == []:
        # reset chat history when user runs clear
        chat = make_chat_session()

    chat.reply(message)
    computed_gen = str(chat.last)
    for i in range(0, len(computed_gen), 8):
        time.sleep(0.05)
        yield computed_gen[: i + 8]
    return 

demo = gr.ChatInterface(fn=respond, 
                        examples=["hello what is the python language?"], 
                        undo_btn=None,
                        retry_btn=None,
                        title="Llama 70b Foundational Model Chat Bot").queue()

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.gradio(demo).run()

# COMMAND ----------





import abc
import datetime
import json
import logging
import os
import threading
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Any, Literal, Optional
from urllib.parse import urlparse

from databricks.sdk import WorkspaceClient

from dbtunnel.utils import pkill, ctx, get_logger, execute


class DBTunnelError(Exception):
    pass


@dataclass
class ProxySettings:
    proxy_url: str
    port: str
    url_base_path: str
    url_base_path_no_port: Optional[str] = None

    def get_proxy_url(self, ensure_ends_with_slash=False):
        """
        For certain apps that use relative paths like "assets/index-*.js" we need to ensure that the url ends
        with a slash.
        """
        if ensure_ends_with_slash is True:
            return self.proxy_url.rstrip("/") + "/"
        return self.proxy_url


def get_cloud(context: Dict[str, Any]) -> str:
    # TODO: support gcp
    if context["extraContext"]["api_url"].endswith("azuredatabricks.net"):
        return "azure"
    return "aws"


def remove_lowest_subdomain_from_host(url):
    parsed_url = urlparse(url)
    host = parsed_url.netloc if parsed_url.netloc else parsed_url.path
    parts = host.split('.')
    # Check if there are subdomains to remove
    if len(parts) > 2:
        # Remove the lowest subdomain
        parts.pop(0)

    # Reconstruct the modified host
    modified_host = '.'.join(parts)

    return modified_host


def get_cloud_proxy_settings(cloud: str, org_id: str, cluster_id: str, port: int) -> ProxySettings:
    cloud_norm = cloud.lower()
    if cloud_norm not in ["aws", "azure"]:
        raise Exception("only supported in aws or azure")
    prefix_url_settings = {
        "aws": "https://dbc-dp-",
        "azure": "https://adb-dp-",
    }
    suffix_url_settings = {
        "azure": "azuredatabricks.net",
    }
    if cloud_norm == "aws":
        suffix = remove_lowest_subdomain_from_host(ctx.host)
        suffix_url_settings["aws"] = suffix

    org_shard = ""
    # org_shard doesnt need a suffix of "." for dnsname its handled in building the url
    # only azure right now does dns sharding
    # gcp will need this
    if cloud_norm == "azure":
        org_shard_id = int(org_id) % 20
        org_shard = f".{org_shard_id}"

    url_base_path_no_port = f"/driver-proxy/o/{org_id}/{cluster_id}"
    url_base_path = f"{url_base_path_no_port}/{port}/"
    return ProxySettings(
        proxy_url=f"{prefix_url_settings[cloud_norm]}{org_id}{org_shard}.{suffix_url_settings[cloud_norm]}{url_base_path}",
        port=str(port),
        url_base_path=url_base_path,
        url_base_path_no_port=url_base_path_no_port
    )


Flavor = Literal[
    "gradio", "fastapi", "nicegui", "streamlit", "stable-diffusion-ui", "bokeh", "flask", "dash", "solara",
    "code-server", "chainlit", "shiny-python", "uvicorn", "arize-phoenix"]


def get_current_username() -> str:
    return ctx.current_user_name


def extract_hostname(url):
    parsed_url = urlparse(url)
    return parsed_url.hostname


def ensure_scheme(url):
    if not url.startswith("http") and not url.startswith("https"):
        return f"https://{url}"


# TODO: Make the with commands lazy so the logger and other
#  init methods are executed first before the with commands

class DbTunnel(abc.ABC):

    def __init__(self, port: int, flavor: Flavor):
        self._port = port
        self._flavor = flavor
        import IPython
        self._dbutils = IPython.get_ipython().user_ns["dbutils"]
        self._display_html = IPython.get_ipython().user_ns["displayHTML"]
        self._context = json.loads(self._dbutils.notebook.entry_point.getDbutils().notebook().getContext().toJson())
        self._org_id = self._context["tags"]["orgId"]
        self._cluster_id = self._context["tags"]["clusterId"]
        # need to do this after the context is set
        self._cloud = get_cloud(self._context)
        self._proxy_settings = get_cloud_proxy_settings(self._cloud, self._org_id, self._cluster_id,
                                                        self._port)
        self._loop = None
        self._share = False
        self._share_information = None
        self._share_trigger_callback = None
        self._log: logging.Logger = get_logger()  # initialize logger during the run method
        self._basic_tunnel_auth = {"token_auth": False, "token_auth_workspace_url": None}

    def _is_single_user_cluster(self):
        ws = WorkspaceClient()
        cluster = ws.clusters.get(self._cluster_id)
        from databricks.sdk.service.compute import DataSecurityMode
        return (cluster.data_security_mode in [DataSecurityMode.SINGLE_USER, DataSecurityMode.LEGACY_SINGLE_USER]) \
            and cluster.single_user_name == get_current_username()

    @abc.abstractmethod
    def _imports(self):
        pass

    @abc.abstractmethod
    def _run(self):
        pass

    @abc.abstractmethod
    def _display_url(self):
        pass

    @property
    def shared(self):
        return self._share

    def inject_auth(self, host: str = None, token: str = None, write_cfg: bool = False):
        """
        Inject databricks host and token into the environment

        :param host: any databricks url you may want to use, otherwise defaults to your notebook session host
        :param token: any databricks token you may want to use, otherwise defaults to your notebook session token
        :param write_cfg: this will create a .databrickscfg file in the root directory, only works on single user clusters.
        """
        if os.getenv("DATABRICKS_HOST") is None:
            self._log.info("Setting databricks host from context")
            os.environ["DATABRICKS_HOST"] = host or ensure_scheme(ctx.host)
        if os.getenv("DATABRICKS_TOKEN") is None:
            self._log.info("Setting databricks token from context")
            os.environ["DATABRICKS_TOKEN"] = token or ctx.token

        if write_cfg is True and self._is_single_user_cluster():
            expanded_file_path = os.path.expanduser("~/.databrickscfg")
            with open(expanded_file_path, "w") as f:
                f.write(f"[DEFAULT]\nhost = {os.getenv('DATABRICKS_HOST')}\ntoken = {os.getenv('DATABRICKS_TOKEN')}\n")

        return self

    def inject_sql_warehouse(self, http_path: str, server_hostname: str = None, token: str = None):
        """
        Inject databricks warehouse http path into the environment and auth
        :param http_path: the http path to the warehouse
        :param server_hostname: the hostname of the databricks workspace, defaults to the current workspace where the notebook is running
        :param token: any databricks token you may want to use, otherwise defaults to your notebook session token
        :return:
        """
        if os.getenv("DATABRICKS_SERVER_HOSTNAME") is None:
            self._log.info("Setting databricks server hostname from context")
            os.environ["DATABRICKS_SERVER_HOSTNAME"] = server_hostname or extract_hostname(
                ctx.host)

        if os.getenv("DATABRICKS_TOKEN") is None:
            self._log.info("Setting databricks token from context")
            os.environ["DATABRICKS_TOKEN"] = token or ctx.token

        if os.getenv("DATABRICKS_HTTP_PATH") is None:
            self._log.info("Setting databricks warehouse http path")
            os.environ["DATABRICKS_HTTP_PATH"] = http_path

        return self

    def inject_env(self, **kwargs):
        """
        Inject environment variables into the environment. Keep in mind environment variables are case sensitive

        Example usage:
        dbtunnel.chainlit("path/to/script").inject_env(ENV_VAR1="value1", ENV_VAR2="value2").run()

        :param kwargs: keyword arguments for the environment variables you want to set
        :return:
        """
        for k, v in kwargs.items():
            if type(v) != str:
                raise ValueError(f"Value for environment variable {k} must be a string")
            self._log.info(f"Setting environment variable {k}")
            os.environ[k] = v
        return self

    def with_token_auth(self):
        """
        Experimental feature do not use.
        :return:
        """

        self._basic_tunnel_auth["token_auth"] = True
        self._basic_tunnel_auth["token_auth_workspace_url"] = ctx.host
        return self

    def with_custom_logger(self, *,
                           logger: Optional[logging.Logger] = None,
                           app_name: str = "dbtunnel",
                           cluster_logging_file_path: Optional[Path] = None,
                           logging_archive_folder: Optional[Path] = None,
                           rotate_when: Literal["S", "M", "H", "D", "midnight"] = "H",
                           rotate_interval: int = 1,
                           backup_count: int = 3,
                           at_time: Optional[datetime.time] = None,
                           format_str: str = "[%(asctime)s] [%(levelname)s] {%(module)s.py:%(funcName)s:%(lineno)d} - %(message)s",
                           datefmt_str: str = "%Y-%m-%dT%H:%M:%S%z"
                           ):
        if logger is not None:
            self._log = logger
            return self
        self._log = get_logger(app_name=app_name,
                               cluster_logging_file_path=cluster_logging_file_path,
                               logging_archive_folder=logging_archive_folder,
                               rotate_when=rotate_when,
                               rotate_interval=rotate_interval,
                               backup_count=backup_count,
                               at_time=at_time,
                               format_str=format_str,
                               datefmt_str=datefmt_str)
        return self

    def _validate_options(self):
        if self._share is True and self._basic_tunnel_auth["token_auth"] is True:
            raise DBTunnelError("Cannot use token auth with shared tunnel; remove token auth or remove sharing")

    def run(self):
        """
        Lifecycle:
        1. initialize logger
        2. import libraries and return error immediately if things are not installed
        3. Then spawn processes.
        :return:
        """
        self._imports()
        self._validate_options()
        if self._share is True and self._share_trigger_callback is not None:
            import nest_asyncio
            nest_asyncio.apply()
            self._share_trigger_callback()
        if self._share is True and self._share_information is not None:
            self._log.info(f"Use this information to publicly access your app: \n{self._share_information.public_url}")
        self._run()

    def share_to_internet(self,
                          *,
                          app_name: str,
                          tunnel_host: str,
                          app_host: str,
                          tunnel_port: int = 7000,
                          subdomain: str = None,
                          sso: bool = False):
        self._share = True
        from dbtunnel.relay import DBTunnelRelayClient
        dbtunnel_relay_client = DBTunnelRelayClient(
            app_name=app_name,
            app_host=app_host,
            tunnel_host=tunnel_host,
            tunnel_port=tunnel_port,
            local_port=self._port,
            subdomain=subdomain,
            sso=sso,
            user=ctx.current_user_name
        )
        print("Downloading required binary if it does not exist!")
        dbtunnel_relay_client.download_on_linux()

        def share_to_internet():
            try:
                print("Access your app at: ", dbtunnel_relay_client.public_url())
                dbtunnel_relay_client.run_as_thread(output_func=self._log.info)
            except Exception as e:
                self._log.error(e)

        self._share_trigger_callback = share_to_internet
        return self

    # right now only ngrok is supported so auth token is required field but in future there may be devtunnels
    def share_to_internet_via_ngrok(self,
                                    *,
                                    ngrok_api_token: str,
                                    ngrok_tunnel_auth_token: str,
                                    kill_existing_processes: bool = True,
                                    kill_all_tunnel_sessions: bool = False,
                                    basic_auth: str = None,  # "databricks:password"
                                    domain: str = None,
                                    oauth_provider: str = None,
                                    oauth_allow_domains: list[str] = None):
        self._share = True
        if kill_existing_processes is True:
            try:
                pkill("ngrok")
            except KeyError:
                self._log.error("no running tunnels to kill")
        from dbtunnel.ngrok import NgrokTunnel
        ngrok_tunnel = NgrokTunnel(self._port,
                                   ngrok_tunnel_auth_token,
                                   ngrok_api_token,
                                   self._log,
                                   basic_auth=basic_auth,
                                   domain=domain,
                                   oauth_provider=oauth_provider,
                                   oauth_allow_domains=oauth_allow_domains)

        def ngrok_callback():
            if kill_all_tunnel_sessions is True:
                ngrok_tunnel.kill_existing_sessions()
            ngrok_tunnel.run()

        self._share_trigger_callback = ngrok_callback

        return self

    def display(self):
        pass  # no op because of the annoying flickering
        # self._display_html(self._display_url())


class DbTunnelProxy:

    def __init__(self,
                 proxy_port: int,
                 service_port: int,
                 url_base_path: str,
                 framework: str,
                 token_auth: bool = False,
                 token_auth_workspace_url: Optional[str] = None,
                 cwd: str = None):
        self._proxy_port = proxy_port
        self._service_port = service_port
        self._url_base_path = url_base_path
        self._framework = framework
        self._token_auth = token_auth
        self._token_auth_workspace_url = token_auth_workspace_url
        self._cwd = cwd
        self._log: logging.Logger = get_logger(app_name="dbtunnel-proxy")
        self._thread = self._make_thread()

    def _make_thread(self):
        my_env = os.environ.copy()

        def run_uvicorn_app(env_copy):
            proxy_cmd = ["python", "-m", "dbtunnel.vendor.asgiproxy",
                         "--port", str(self._proxy_port),
                         "--service-port", str(self._service_port),
                         "--url-base-path", self._url_base_path,
                         "--framework", self._framework]
            if self._token_auth is True:
                proxy_cmd.append("--token-auth")
            if self._token_auth_workspace_url is not None:
                proxy_cmd.append("--token-auth-workspace-url")
                proxy_cmd.append(self._token_auth_workspace_url)

            self._log.info(f"Running proxy server via command: {' '.join(proxy_cmd)}")
            try:
                for log_stmt in execute(proxy_cmd, env_copy, cwd=self._cwd):
                    self._log.info(log_stmt.rstrip("\n"))
            except Exception as e:
                self._log.info("Error running proxy server")

        return threading.Thread(target=run_uvicorn_app, args=(my_env,))

    def start(self):
        self._thread.start()
        return self

    def wait(self):
        self._thread.join()
        return self



import subprocess

from dbtunnel.utils import ctx, compute_utils


class AppTunnels:
    """
    This class is used to create tunnels to different apps. It is the user facing api should not have breaking changes.
    """

    @staticmethod
    def kill_port(port: int):
        subprocess.run(f"kill -9 $(lsof -t -i:{port})", capture_output=True, shell=True)

    @staticmethod
    def fastapi(app, port: int = 8080):
        from dbtunnel.fastapi import FastApiAppTunnel
        return FastApiAppTunnel(app, port)

    @staticmethod
    def gradio(app=None,
               path: str = None,
               cwd: str = None,
               port: int = 8080):
        if app is None and path is None:
            raise ValueError("Either gradio app object or path must be provided")
        from dbtunnel.gradio import GradioAppTunnel
        return GradioAppTunnel(app, path, cwd, port)

    @staticmethod
    def nicegui(app, storage_secret: str = "", port: int = 8080):
        from dbtunnel.nicegui import NiceGuiAppTunnel
        return NiceGuiAppTunnel(app, storage_secret, port)

    @staticmethod
    def streamlit(path, port: int = 8080):
        from dbtunnel.streamlit import StreamlitTunnel
        return StreamlitTunnel(path, port)

    @staticmethod
    def stable_diffusion_ui(no_gpu: bool, port: int = 7860, enable_insecure_extensions: bool = False,
                            extra_flags: str = ""):
        # todo auto detect with torch
        from dbtunnel.stable_diffusion_ui import StableDiffusionUITunnel
        return StableDiffusionUITunnel(no_gpu, port, enable_insecure_extensions, extra_flags)

    @staticmethod
    def bokeh(path, port: int = 8080):
        from dbtunnel.bokeh import BokehTunnel
        return BokehTunnel(path, port)

    @staticmethod
    def flask(app, port: int = 8080):
        from dbtunnel.flask import FlaskAppTunnel
        return FlaskAppTunnel(app, port)

    @staticmethod
    def dash(app, port: int = 8080):
        from dbtunnel.dash import DashAppTunnel
        return DashAppTunnel(app, port)

    @staticmethod
    def solara(path, port: int = 8080):
        from dbtunnel.solara import SolaraAppTunnel
        return SolaraAppTunnel(path, port)

    @staticmethod
    def code_server(directory_path: str = None, repo_name: str = None, port: int = 9988,
                    extension_ids: list[str] = None):
        from dbtunnel.code_server import CodeServerTunnel
        return CodeServerTunnel(directory_path, repo_name, port, extension_ids)

    @staticmethod
    def chainlit(script_path: str, cwd: str = None, port: int = 8000):
        from dbtunnel.chainlit import ChainlitAppTunnel
        return ChainlitAppTunnel(script_path, cwd=cwd, port=port)

    @staticmethod
    def shiny_python(app, port: int = 8080):
        from dbtunnel.shiny import ShinyPythonAppTunnel
        return ShinyPythonAppTunnel(app, port)

    @staticmethod
    def uvicorn(app, port: int = 8080):
        from dbtunnel.uvicorn import UvicornAppTunnel
        return UvicornAppTunnel(app, port)

    @staticmethod
    def arize_phoenix(port: int = 9098):
        from dbtunnel.arize_phoenix_ui import ArizePhoenixUITunnel
        return ArizePhoenixUITunnel(port)


dbtunnel = AppTunnels()



import argparse
import asyncio
from typing import Tuple
from urllib.parse import urlparse

from starlette.types import ASGIApp

from dbtunnel.vendor.asgiproxy.config import BaseURLProxyConfigMixin, ProxyConfig
from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.frameworks import framework_specific_proxy_config
from dbtunnel.vendor.asgiproxy.simple_proxy import make_simple_proxy_app

try:
    import uvicorn
except ImportError:
    uvicorn = None


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("--port", type=int, required=True)
    ap.add_argument("--service-port", type=int, required=True)
    ap.add_argument("--token-auth", action='store_true', default=False)
    ap.add_argument("--token-auth-workspace-url", type=str, default=None)
    ap.add_argument("--host", type=str, default="0.0.0.0")
    ap.add_argument("--url-base-path", type=str, required=True)
    ap.add_argument("--framework", type=str, required=True)
    args = ap.parse_args()
    if not uvicorn:
        ap.error(
            "The `uvicorn` ASGI server package is required for the command line client."
        )
    print("Starting proxy server... with args: ", args)
    config = framework_specific_proxy_config[args.framework](**{
        "url_base_path": args.url_base_path,
        "service_host": args.host,
        "service_port": args.service_port,
        "auth_config": {"token_auth": args.token_auth, "token_auth_workspace_url": args.token_auth_workspace_url}
    })
    proxy_context = ProxyContext(config)
    app = make_simple_proxy_app(proxy_context, framework=args.framework, proxy_port=args.port)
    try:
        return uvicorn.run(host=args.host,
                           port=int(args.port),
                           app=app,
                           root_path=args.url_base_path)
    finally:
        asyncio.run(proxy_context.close())


if __name__ == "__main__":
    main()



import functools
import os
import secrets
import socket
import subprocess
import webbrowser
from pathlib import Path

import click
from click import ClickException

from dbtunnel.relay import DBTunnelRelayClient, ProxyWithNameAlreadyExists, is_url_friendly, TunnelConfigError, \
    StandardProxyError


@click.group()
def cli():
    """
    DBTunnel CLI to poke a hole to internet. This CLI is only for internal use.
    """

    pass


def validate_app_name(ctx, param, value):
    if value is None:
        return value
    if is_url_friendly(value) is False:
        raise click.BadParameter('App name should only contain alphanumeric characters and no spaces.')
    return value


def is_frpc_installed(frpc_path: Path, silent=False):
    stdout_str = None
    stderr_str = None
    try:
        output = subprocess.run([str(frpc_path), "--help"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                env=os.environ)
        stdout_str = output.stdout.decode("utf-8").strip()
        stderr_str = output.stderr.decode("utf-8").strip()
        if silent is False:
            click.echo("✅  frpc is already installed.")
        return True
    except FileNotFoundError:
        click.clear()
        click.echo("❌  frpc is not installed.\n")
        raise click.ClickException(
            "❌  Please install frpc using brew. Run the following line: \n\nHOMEBREW_NO_AUTO_UPDATE=1 brew install frpc\n\n\n\n\n")
    except subprocess.CalledProcessError:
        click.echo("❌  uhoh something happened.")
        click.echo(f"stdout: {stdout_str}")
        click.echo(f"stderr: {stderr_str}")
        return False


def get_frpc_homebrew_path() -> Path:
    try:
        result = subprocess.run(["brew", "--prefix"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                env=os.environ)
        output = result.stdout.decode("utf-8").strip()
        return Path(output) / "bin" / "frpc"
    except subprocess.CalledProcessError:
        raise ValueError("Error: unable to find brew --prefix.")


def verify_homebrew():
    try:
        subprocess.run(["brew", "--help"], check=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                       env=os.environ)
        click.echo("✅  Homebrew is installed.")
        return True
    except Exception:
        click.clear()
        click.echo("❌  Homebrew is not installed.\n")
        raise click.ClickException(
            "❌  Homebrew is not installed. Please install homebrew (https://brew.sh/) and try again.")


def is_mac():
    import platform
    return platform.system() == 'Darwin'


def verify_installation(frpc_exec_path: Path):
    if is_mac() is False:
        raise ClickException("Error: Unsupported platform. Only MacOS is supported for native connection.")
    if is_frpc_installed(frpc_exec_path) is False:
        # just double verify
        is_frpc_installed(frpc_exec_path)
        click.echo(f"✅  frpc is installed at path: {frpc_exec_path}.")


@cli.command()
def validate():
    """
    Validate if frpc is installed and working.
    """
    click.echo('Checking if binaries is installed')
    verify_homebrew()
    frpc_path = get_frpc_homebrew_path()
    verify_installation(frpc_path)
    click.echo('✅  Binaries Properly Installed')


def make_fake_name():
    login = os.getlogin()
    # replace . with - and remove all other special characters
    new_login = login.lower().replace(".", "-")
    return "".join([c for c in new_login if c.isalnum() or c == "-"])


def is_interactive():
    return os.isatty(0)


def open_browser(url):
    webbrowser.open_new_tab(url)


def make_visitor_command_string(app_name, secret_value):
    return f"dbtunnel visit -n {app_name} -s {secret_value}"


def native_flow():
    click.echo('Checking if binaries is installed')
    verify_homebrew()
    exec_path = get_frpc_homebrew_path()
    verify_installation(exec_path)
    click.echo('✅  Binaries Properly Installed')
    return exec_path


def find_next_open_port(*, host="0.0.0.0", start_port: int = 9000, end_port: int = 9999):
    for port in range(start_port, end_port + 1):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)  # Set a timeout for the connection attempt
        result = sock.connect_ex((host, port))
        if result != 0:  # 0 means the port is open
            return port
        sock.close()

    return None


@cli.command()
@click.option("--headless", "-h", is_flag=True, show_default=True, default=False, help="Run without opening browser.")
@click.option('--tunnel-host', '-th', type=str, required=True,
              help='The domain of the dbtunnel server')
@click.option('--tunnel-port', '-tp', type=int, show_default=True,
              default=7000,
              help='The port of the dbtunnel server.')
@click.option('--local-host', type=str, default="0.0.0.0", help='The local host to bind')
@click.option('--local-port', '-p', type=int, default=find_next_open_port(), help='The local port to bind')
@click.option('--app-name', '-n', type=str, default=make_fake_name(), show_default=True, callback=validate_app_name,
              help='The name of the app. Should be unique no spaces.')
@click.option("--secret-value", "-s", show_default=True, default=lambda: secrets.token_urlsafe(32),
              help="Visitor secret value.")
def visit(**kwargs):
    app_name = kwargs.get('app_name')
    tunnel_host = kwargs.get('tunnel_host').replace("https://", "").replace("https://", "").replace("/", "")
    tunnel_port = kwargs.get('tunnel_port')
    local_port = kwargs.get('local_port')
    local_host = kwargs.get('local_host')
    debug_mode = kwargs.get('debug')
    headless = kwargs.get('headless')
    secret_value = kwargs.get('secret_value')

    frpc_native_executable_path = native_flow()

    try:
        db_tunnel_relay_client = DBTunnelRelayClient(
            app_name=app_name,
            tunnel_host=tunnel_host,
            tunnel_port=tunnel_port,
            local_port=local_port,
            local_host=local_host,
            subdomain=app_name,
            executable_path=str(frpc_native_executable_path),
            mode='native',
            log_level="debug" if debug_mode else "info",
            secret=True,
            secret_string=secret_value,
            visitor=True
        )
        click.clear()
        click.echo(click.style(f'DBTunnel by @stikkireddy\n', fg='green', bold=True))
        click.echo(f'💼 Welcome to: {click.style(app_name, fg="green")} | '
                   f'Type: Visiting App via TCP Tunnel \n')
        local_url = f'http://{local_host}:{local_port}'
        styled_local_url = f"\033]8;;{local_url}\033\\{local_url}\033]8;;\033\\"
        styled_local_url = click.style(styled_local_url, fg="cyan")
        click.echo(f'🚀 Local URL: {styled_local_url}\n')

        open_browser_f = functools.partial(open_browser, local_url)
        click.echo(click.style('Tunnel Logs: ', fg='green', bold=True))
        db_tunnel_relay_client.run(output_func=click.echo,
                                   success_callback=open_browser_f if headless is False else None)
    except (ProxyWithNameAlreadyExists, StandardProxyError) as e:
        raise click.ClickException(str(e))
    except TunnelConfigError as e:
        raise click.ClickException(f"Error Configuring Tunnel: {str(e)}")
    except subprocess.CalledProcessError as e:
        click.secho("\n\n❌  uhoh something happened.", fg="red", bold=True)
        styled_error = click.style(f"if you get an error indicating that the proxy already exists please use -n to "
                                   f"provide a unique name", fg="red", bold=True)
        raise click.ClickException(styled_error)


@cli.command()
@click.option("--headless", "-h", is_flag=True, show_default=True, default=False, help="Run without opening browser.")
@click.option("--debug", is_flag=True, show_default=True, default=False, help="Enable debug mode.")
@click.option('--tunnel-host', '-th', type=str, required=True, help='The domain of the dbtunnel server')
@click.option('--app-host', '-ah', type=str, required=True, help='The domain where the app is going to be served from')
@click.option('--tunnel-port', '-tp', type=int, show_default=True,
              default=7000,
              help='The port of the dbtunnel server.')
@click.option('--local-host', type=str, default="0.0.0.0", help='The local host to bind')
@click.option('--local-port', '-p', type=int, required=True, help='The local port to bind')
@click.option('--app-name', '-n', type=str, default=make_fake_name(), show_default=True, callback=validate_app_name,
              help='The name of the app. Should be unique no spaces.')
@click.option('--subdomain', '-sd', type=str, callback=validate_app_name,
              help='The subdomain of the app. Should be unique no spaces.')
@click.option("--sso", is_flag=True, show_default=True, default=False,
              help="Causes the tunnel to require sso via cloudflare zero trust and secured as setup by WAF rules.")
@click.option("--ssh", is_flag=True, show_default=True, default=False,
              help="Use ssh instead of native cli for tunnel.")
@click.option("--native", is_flag=True, show_default=True, default=False,
              help="Use native instead of ssh for tunnel.")
@click.option("--secret", is_flag=True, show_default=True, default=False,
              help="Entirely private connection, no public urls.")
@click.option("--secret-value", show_default=True, default=lambda: secrets.token_urlsafe(32),
              help="Private key for your visitors.")
def bind(**kwargs):
    """
    Bind a local port to a dbtunnel server domain.
    """
    use_ssh_mode = kwargs.get('ssh')
    use_native_mode = kwargs.get('native')
    use_best_mode = use_ssh_mode is False and use_native_mode is False
    app_name = kwargs.get('app_name')
    tunnel_host = kwargs.get('tunnel_host').replace("https://", "").replace("https://", "").replace("/", "")
    tunnel_port = kwargs.get('tunnel_port')
    app_host = kwargs.get('app_host').replace("https://", "").replace("https://", "").replace("/", "")
    local_port = kwargs.get('local_port')
    local_host = kwargs.get('local_host')
    sso = kwargs.get('sso')
    debug_mode = kwargs.get('debug')
    headless = kwargs.get('headless')
    secret = kwargs.get('secret')
    secret_value = kwargs.get('secret_value')
    frpc_native_executable_path = None

    if use_native_mode is True:
        # usually you have ssh client installed on mac
        frpc_native_executable_path = native_flow()

    if use_best_mode is True:
        try:
            frpc_native_executable_path = native_flow()
        except ClickException:
            click.echo("❌  Native mode failed. Trying ssh tunnel instead.")
            use_ssh_mode = True
            del use_native_mode

    click.echo(f'✅  Pushing {app_name} to dbtunnel server via: {"native" if use_ssh_mode is False else "ssh tunnel"}')
    try:
        db_tunnel_relay_client = DBTunnelRelayClient(
            app_name=app_name,
            tunnel_host=tunnel_host,
            tunnel_port=tunnel_port,
            local_port=local_port,
            app_host=app_host,
            subdomain=app_name,
            executable_path=str(frpc_native_executable_path),
            mode='ssh' if use_ssh_mode else 'native',
            sso=sso,
            log_level="debug" if debug_mode else "info",
            secret=secret,
            secret_string=secret_value
        )

        click.clear()
        click.echo(click.style(f'DBTunnel by @stikkireddy\n', fg='green', bold=True))
        click.echo(f'💼 App Name: {click.style(app_name, fg="green")} | '
                   f'Type: {click.style("TCP over SSH Tunnel" if use_ssh_mode is True else "TCP Tunnel", fg="green")}\n')
        local_url = f'http://{local_host}:{local_port}'
        local_url = f"\033]8;;{local_url}\033\\{local_url}\033]8;;\033\\"
        styled_local_url = click.style(local_url, fg="cyan")
        click.echo(f'🚀 APP URL: {styled_local_url}\n')

        open_browser_f = None

        # public url is only for non secret apps
        if secret is False:
            url = db_tunnel_relay_client.public_url()
            hyper_link = f'\033]8;;{url}\033\\{url}\033]8;;\033\\'
            click.echo(
                f'🔗 {"SSO Secured" if sso else "Public Sharable"} URL: {click.style(hyper_link, fg="green")}\n', )
            open_browser_f = functools.partial(open_browser, url)
        else:
            click.echo(f'🔒 Visitor Command: '
                       f'{click.style(make_visitor_command_string(app_name, secret_value), fg="green")}\n')
        click.echo(click.style('Tunnel Logs: ', fg='green', bold=True))
        db_tunnel_relay_client.run(output_func=click.echo,
                                   success_callback=open_browser_f if is_interactive() and headless is False else None)
    except (ProxyWithNameAlreadyExists, StandardProxyError) as e:
        raise click.ClickException(str(e))
    except TunnelConfigError as e:
        raise click.ClickException(f"Error Configuring Tunnel: {str(e)}")
    except subprocess.CalledProcessError as e:
        click.secho("\n\n❌  uhoh something happened.", fg="red", bold=True)
        styled_error = click.style(f"if you get an error indicating that the proxy already exists please use -n to "
                                   f"provide a unique name", fg="red", bold=True)
        raise click.ClickException(styled_error)



import functools
from dataclasses import dataclass
from pathlib import Path

from enum import Enum

from cachetools import TTLCache
from databricks.sdk import WorkspaceClient
from starlette.requests import Request
from starlette.responses import Response, RedirectResponse
from starlette.types import ASGIApp, Receive, Scope, Send

from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.proxies.http import proxy_http
from dbtunnel.vendor.asgiproxy.proxies.websocket import proxy_websocket
from dbtunnel.vendor.asgiproxy.utils.headers import add_if_databricks_proxy_scope, is_from_databricks_proxy, \
    add_framework_to_scope, add_origin_port_to_scope

DB_TUNNEL_LOGIN_PATH = "/dbtunnel/login"


@functools.lru_cache(maxsize=0)
def get_login_content(*,
                      root_path: str,
                      workspace_url: str,
                      user_name: str):
    from dbtunnel.vendor.asgiproxy import templates
    templates_directory = Path(templates.__file__).parent

    # Construct the path to the login.html file
    login_html_path = templates_directory / 'login.html'

    with open(str(login_html_path), "r") as file:
        # root path needs right slash trimmed off
        return file.read().format(workspace_url=workspace_url,
                                  root_path=root_path.rstrip("/"),
                                  login_path=DB_TUNNEL_LOGIN_PATH,
                                  user_name=user_name)


@dataclass
class DatabricksProxyHeaders:
    user_id: str
    user_name: str


def get_databricks_user_header(scope: Scope) -> DatabricksProxyHeaders:
    hdrs = DatabricksProxyHeaders(user_id="", user_name="")
    for header in scope["headers"]:
        key = header[0].decode("utf-8")
        if key.lower() == "x-databricks-user-name":
            hdrs.user_name = header[1].decode("utf-8")
        if key.lower() == "x-databricks-user-id":
            hdrs.user_id = header[1].decode("utf-8")
    return hdrs


def validate_user(url: str, user: str, token: str) -> bool:
    try:
        w = WorkspaceClient(
            host=url,
            token=token
        )
        return w.current_user.me().user_name == user
    except Exception:
        return False


# Simple state machine, user is not in authloop when ttl cache has their email with token
class AuthLoopState(Enum):
    StillInAuthLoop = "STILL_IN_AUTH_LOOP"
    NotInAuthLoop = "NOT_IN_AUTH_LOOP"


async def handle_token_auth(
        proxy_context: ProxyContext,
        cache: TTLCache,
        scope: Scope,
        send: Send,
        receive: Receive):
    workspace_url = proxy_context.config.token_auth_workspace_url
    if is_from_databricks_proxy(scope) is True:
        root_path = scope["root_path"]
        if scope["path"].startswith(root_path):
            scope["path"] = scope["path"].replace(root_path, "")
    else:
        root_path = "/"
    dbx_ctx_headers = get_databricks_user_header(scope)
    login_page_content = get_login_content(
        workspace_url=workspace_url,
        user_name=dbx_ctx_headers.user_name,
        root_path=root_path
    )

    if scope["path"] == DB_TUNNEL_LOGIN_PATH:
        request = Request(scope, receive)
        content = await request.form()
        user = content.get("userName") or dbx_ctx_headers.user_name
        workspace_url = content.get("workspaceUrl") or workspace_url
        password = content.get("token")
        # user accidentally clicks form
        if user is None and password is None:
            # this is root path with respect to asgi app root path
            resp = RedirectResponse(url="/", status_code=302)
            await resp(scope, receive, send)
            return AuthLoopState.StillInAuthLoop
        # invalid user
        if validate_user(workspace_url, user, password) is False:
            resp = Response(
                content=login_page_content, status_code=401
            )
            await resp(scope, receive, send)
            return AuthLoopState.StillInAuthLoop

        cache[user] = password
        # this is root path with respect to asgi app root path
        resp = RedirectResponse(url="/", status_code=302)
        await resp(scope, receive, send)
        return AuthLoopState.StillInAuthLoop

    # TODO add fernet encryption
    token = cache.get(dbx_ctx_headers.user_name)
    if token is None:
        resp = Response(
            content=login_page_content, status_code=200
        )
        await resp(scope, receive, send)
        return AuthLoopState.StillInAuthLoop

    if token is not None:
        scope["headers"].append((b"Authorization", f"Bearer {token}".encode("utf-8")))

    return AuthLoopState.NotInAuthLoop


def make_simple_proxy_app(
        proxy_context: ProxyContext,
        framework: str,
        proxy_port: int,
        login_timeout: int = 3600,
        *,
        proxy_http_handler=proxy_http,
        proxy_websocket_handler=proxy_websocket,
) -> ASGIApp:
    """
    Given a ProxyContext, return a simple ASGI application that can proxy
    HTTP and WebSocket connections.

    The handlers for the protocols can be overridden and/or removed with the
    respective parameters.
    """

    # we assume there is not going to be more than 250k users
    cache = TTLCache(maxsize=250000, ttl=login_timeout)  # just dont use it if auth not needed

    async def app(scope: Scope, receive: Receive, send: Send):  # noqa: ANN201

        if scope["type"] == "lifespan":
            return None  # We explicitly do nothing here for this simple app.

        add_framework_to_scope(scope, framework)
        add_if_databricks_proxy_scope(scope)
        add_origin_port_to_scope(scope, proxy_port)
        if is_from_databricks_proxy(scope) is False:
            # remove all the driver proxy defaults this is usually when it comes from a relay/etc
            new_root_path = "/"
            current_root_path = scope["root_path"]
            if scope["path"].startswith(current_root_path):
                scope["path"] = scope["path"].replace(current_root_path, "")
            scope["root_path"] = new_root_path

        # we do not have enough information in websocket proxied headers to function auth
        if proxy_context.config.token_auth_workspace_url is not None and proxy_context.config.token_auth is True and \
                scope["type"] == "http":
            resp = await handle_token_auth(proxy_context, cache, scope, send, receive)
            if resp == AuthLoopState.StillInAuthLoop:
                return None

        if scope["type"] == "http" and proxy_http_handler:
            return await proxy_http_handler(
                context=proxy_context, scope=scope, receive=receive, send=send
            )

        if scope["type"] == "websocket" and proxy_websocket_handler:
            return await proxy_websocket_handler(
                context=proxy_context, scope=scope, receive=receive, send=send
            )

        raise NotImplementedError(f"Scope {scope} is not understood")

    return app



import hashlib
import os
import re
import secrets
import shutil
import tarfile
import tempfile
import uuid
from typing import Optional, Literal, List

import requests

from dbtunnel.utils import execute

PRIVATE_SUBDOMAIN_PREFIX = "private-"


class ProxyWithNameAlreadyExists(Exception):
    pass

class StandardProxyError(Exception):
    pass

class TunnelConfigError(Exception):
    pass


def is_url_friendly(subdomain: str):
    try:
        return all(c.isalnum() or c in "-_" for c in subdomain)
    except ValueError:
        return False


def download_and_copy(url, destination_dir):
    # Create a temporary directory
    frpc_dest_path = os.path.join(destination_dir, 'frpc')
    if os.path.exists(frpc_dest_path):
        print(f"'frpc' already exists in {destination_dir}. Skipping download.")
        return

    with tempfile.TemporaryDirectory() as temp_dir:
        # Download the file
        response = requests.get(url, stream=True)
        file_name = url.split("/")[-1]
        temp_file_path = os.path.join(temp_dir, file_name)

        with open(temp_file_path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        # Extract the tar.gz file
        with tarfile.open(temp_file_path, 'r:gz') as tar:
            tar.extractall(path=temp_dir)

        # Copy the frpc file to the destination directory
        frpc_file_path = os.path.join(temp_dir, 'frp_0.54.0_linux_amd64', 'frpc')
        shutil.copy(frpc_file_path, destination_dir)


class DBTunnelRelaySecretTunnelCfg:

    # only used for more complex settings
    def __init__(self, *,
                 app_name,
                 visitor: bool,
                 tunnel_host="proxy.dbtunnel.app",
                 tunnel_port: int = 7000,
                 tunnel_type: Literal['stcp'] = 'stcp',  # secure tcp
                 server_secret: str = None,
                 bind_port: int = 4040,
                 bind_host: str = "0.0.0.0",
                 folder_name: str = ".dbtunnel", ):
        self._app_name = app_name
        self._tunnel_host = tunnel_host
        self._tunnel_port = tunnel_port
        self._tunnel_type = tunnel_type
        self._visitor = visitor
        if visitor is True and server_secret is None:
            raise ValueError(f"server_secret is required for visiting app: {app_name}")
        if visitor is False and server_secret is None:
            self._server_secret = secrets.token_urlsafe(32)
        else:
            self._server_secret = server_secret
        self._bind_port = bind_port
        self._bind_host = bind_host
        self._folder_name = folder_name
        self._setup()

    def get_cfg_path(self):
        folder_path = os.path.join(os.getcwd(), self._folder_name)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        # Create a file with syntax <visitor|secretserver>.app_name.toml
        file_name = f"{'visitor' if self._visitor is True else 'secretserver'}.{self._app_name}.toml"
        file_path = os.path.join(folder_path, file_name)
        return file_path

    def write_visitor_cfg(self, file_path: str):
        with open(file_path, "w") as f:
            f.write(f"serverAddr = \"{self._tunnel_host}\"\n")
            f.write(f"serverPort = {self._tunnel_port}\n")
            f.write(f"\n")
            f.write(f"[[visitors]]\n")
            f.write(f"name = \"{self._app_name}-visitor\"\n")
            f.write(f"type = \"{self._tunnel_type}\"\n")
            f.write(f"serverName = \"{self._app_name}\"\n")
            f.write(f"secretKey = \"{self._server_secret}\"\n")
            f.write(f"bindAddr = \"{self._bind_host}\"\n")
            f.write(f"bindPort = {self._bind_port}\n")

    def write_secret_server_cfg(self, file_path: str):
        with open(file_path, "w") as f:
            f.write(f"serverAddr = \"{self._tunnel_host}\"\n")
            f.write(f"serverPort = {self._tunnel_port}\n")
            f.write(f"\n")
            f.write(f"[[proxies]]\n")
            f.write(f"name = \"{self._app_name}\"\n")
            f.write(f"type = \"{self._tunnel_type}\"\n")
            f.write(f"secretKey = \"{self._server_secret}\"\n")
            f.write(f"localIP = \"{self._bind_host}\"\n")
            f.write(f"localPort = {self._bind_port}\n")

    def _setup(self):
        # Create .dbtunnel folder if it doesn't exist
        file_path = self.get_cfg_path()
        if self._visitor:
            self.write_visitor_cfg(file_path)
        else:
            self.write_secret_server_cfg(file_path)

    def get_cmd(self, executable: str) -> List[str]:
        return [executable, "-c", self.get_cfg_path()]


# TODO: break this class into more composable components
class DBTunnelRelayClient:
    def __init__(self, *,
                 local_port,
                 app_name,
                 tunnel_host,
                 log_level="info",
                 user: Optional[str] = None,
                 tunnel_port: int = 7000,
                 app_host: Optional[str] = None,
                 subdomain: Optional[str] = None,
                 local_host="0.0.0.0",
                 file_name: str = "dbtunnel.toml",
                 folder_name: str = ".dbtunnel",
                 executable_path: str = "frpc",
                 sso: bool = False,
                 mode: Literal['native', 'ssh'] = 'native',
                 secret: bool = False,
                 visitor: bool = False,
                 secret_string: Optional[str] = None,
                 ):
        # visitor requires secret to be true
        self._secret_string = secret_string
        self._visitor = visitor
        self._secret = secret
        self._log_level = log_level
        self._app_host = app_host
        self._sso = sso
        self._mode = mode
        self._app_name = app_name
        self._tunnel_host = tunnel_host
        self._tunnel_port = tunnel_port
        self._local_port = local_port
        self._local_host = local_host
        self._subdomain: str = self._handle_subdomain(subdomain or self._app_name or str(uuid.uuid4()), sso)
        self._file_name = file_name
        self._folder_name = folder_name
        self._executable_path = executable_path if isinstance(executable_path, str) else str(executable_path)
        self._user = user or os.getlogin().lower()
        self._setup()

    @staticmethod
    def _handle_subdomain(subdomain, is_sso):
        # ensure that the subdomain is url friendly
        if is_url_friendly(subdomain) is False:
            raise TunnelConfigError(f"App name {subdomain} should be alphanumeric ")

        if is_sso is True and subdomain.startswith(PRIVATE_SUBDOMAIN_PREFIX) is False:
            return f"{PRIVATE_SUBDOMAIN_PREFIX}{subdomain}"

        if is_sso is True and subdomain.startswith(PRIVATE_SUBDOMAIN_PREFIX) is True:
            raise TunnelConfigError(f"App name {subdomain} is already private please remove the private- prefix")

        if is_sso is False and subdomain.startswith(PRIVATE_SUBDOMAIN_PREFIX) is True:
            raise TunnelConfigError(f"App name {subdomain} is private please add make sure you"
                                    f" create the tunnel in private mode")

        return subdomain

    def get_file_path(self):
        return os.path.join(os.getcwd(), self._folder_name, self._file_name)

    def download_on_linux(self):
        url = "https://github.com/fatedier/frp/releases/download/v0.54.0/frp_0.54.0_linux_amd64.tar.gz"
        download_and_copy(url, os.path.join(os.getcwd(), self._folder_name))
        self._executable_path = os.path.join(os.getcwd(), self._folder_name, "frpc")

    def _setup(self):
        # Create .dbtunnel folder if it doesn't exist
        folder_path = os.path.join(os.getcwd(), self._folder_name)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

    def validate(self):
        if is_url_friendly(self._subdomain) is False:
            raise ValueError(
                f"Subdomain: {self._subdomain} should be url friendly. Only alphanumeric, hyphen and "
                f"underscore are allowed.")

    def _sanitize_log(self, log_stmt):
        if self._mode == "native":
            # replace [.*\.go.*] with dbtunnel
            return re.sub(r"\[[^\[]*\.go:\d+\]", "[dbtunnel-client]", log_stmt)
        if self._mode == "ssh":
            if "frp (via SSH)" in log_stmt:
                return None
            if log_stmt.strip() == "":
                return None
            if log_stmt.startswith("RemoteAddress"):
                return f"RemoteAddress: {self.public_url()}"
            return log_stmt

    def _get_cmd(self):
        if self._mode == 'native':
            # return [self._executable_path, "-c", self.get_file_path()]
            cmd = [self._executable_path,
                   "http",
                   "--server-addr",
                   self._tunnel_host,
                   "--server-port",
                   str(self._tunnel_port),
                   "--sd",
                   self._subdomain,
                   "-n",
                   self._app_name,
                   "--local-port",
                   str(self._local_port),
                   "--local-ip",
                   self._local_host,
                   "--ue",
                   "--uc",
                   "--log-level",
                   self._log_level,
                   "--user",
                   self._user]
            if any([item is None or isinstance(item, str) is False for item in cmd]):
                raise TunnelConfigError(f"Invalid command: {cmd}")
            return cmd

        else:
            return ["ssh", "-R",
                    f":8000:{self._local_host}:{self._local_port}",
                    f"v0@{self._tunnel_host}", "-p",
                    "7200", "http",
                    "--proxy_name", self._app_name,
                    "--sd", self._subdomain,
                    "-u", self._user]

    @staticmethod
    def has_relay_conn_started(log_stmt: str):
        if "start proxy success" in log_stmt:
            return True
        if "start visitor success" in log_stmt:
            return True
        if "RemoteAddress" in log_stmt:
            return True
        return False

    @staticmethod
    def is_visitor_connection_error(stmt):
        return "start new visitor connection error" in stmt

    def _run(self, cmd: List[str], output_func=None, success_callback=None) -> None:
        r = re.compile(r".*start error: proxy.*already exists.*")
        env_copy = os.environ.copy()
        for stmt in execute(cmd, env=env_copy):
            stmt = self._sanitize_log(stmt)
            if stmt is None:
                continue
            if self.has_relay_conn_started(stmt) and success_callback is not None:
                success_callback()
            if r.match(stmt):
                output_func(stmt.rstrip("\n"))
                raise ProxyWithNameAlreadyExists(f"Proxy [{self._app_name}] already exists."
                                                 f" Please use a different app name.")
            if self.is_visitor_connection_error(stmt):
                output_func(stmt.rstrip("\n"))
                raise StandardProxyError(f"Visitor connection error. Please check with provider if your server is still up.")
            output_func(stmt.rstrip("\n"))

    def get_secret_cmd(self) -> List[str]:
        self.validate()
        if self._mode == 'ssh':
            raise ValueError("Secret tunnel is not supported in ssh mode")
        cfg = DBTunnelRelaySecretTunnelCfg(
            app_name=self._app_name,
            visitor=self._visitor,
            server_secret=self._secret_string,
            tunnel_host=self._tunnel_host,
            tunnel_port=self._tunnel_port,
            tunnel_type="stcp",
            bind_port=self._local_port,
            bind_host=self._local_host,
            folder_name=self._folder_name
        )
        return cfg.get_cmd(self._executable_path)

    def run(self, output_func=None, success_callback=None) -> None:
        self.validate()
        output_func = output_func or print
        # Run the frpc command
        if self._secret is False:
            cmd = self._get_cmd()
            self._run(cmd, output_func, success_callback)
        else:
            cmd = self.get_secret_cmd()
            self._run(cmd, output_func, success_callback)

    def run_as_thread(self, output_func=None, success_callback=None):
        import threading
        t = threading.Thread(target=self.run, args=(output_func, success_callback,))
        t.start()
        return t

    def public_url(self):
        return f"https://{self._subdomain}.{self._app_host}"



# Databricks notebook source
# MAGIC %pip install dbtunnel[bokeh]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/bokeh_sample.py"

# COMMAND ----------

from dbtunnel import dbtunnel

dbtunnel.bokeh(script_path).run()



# Databricks notebook source
# MAGIC %pip install dbtunnel[fastapi]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from fastapi import FastAPI
from fastapi.responses import HTMLResponse

app = FastAPI()


@app.get("/", response_class=HTMLResponse)
def read_root():
    return """
    <html>
        <head>
            <link rel="stylesheet" href="/static/style.css">
        </head>
        <body>
            <h1>Hello, FastAPI!</h1>
            <button onclick="alert('Button Clicked!')">Click me</button>
        </body>
    </html>
    """

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.fastapi(app).run()

# COMMAND ----------





import asyncio
import fnmatch
from typing import AsyncGenerator, Union

import aiohttp
from starlette.requests import Request
from starlette.responses import Response, StreamingResponse
from starlette.types import Receive, Scope, Send

from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.utils.headers import is_from_databricks_proxy
from dbtunnel.vendor.asgiproxy.utils.streams import read_stream_in_chunks

# TODO: make these configurable?
INCOMING_STREAMING_THRESHOLD = 512 * 1024
OUTGOING_STREAMING_THRESHOLD = 1024 * 1024 * 5


def determine_incoming_streaming(request: Request) -> bool:
    if request.method in ("GET", "HEAD"):
        return False

    try:
        return int(request.headers["content-length"]) < INCOMING_STREAMING_THRESHOLD
    except (TypeError, ValueError, KeyError):
        # Malformed or missing content-length header; assume a very large payload
        return True


def determine_outgoing_streaming(proxy_response: aiohttp.ClientResponse) -> bool:
    if proxy_response.status != 200:
        return False
    try:
        return (
                int(proxy_response.headers["content-length"]) > OUTGOING_STREAMING_THRESHOLD
        )
    except (TypeError, ValueError, KeyError):
        # Malformed or missing content-length header; assume a streaming payload
        return True


async def get_proxy_response(
        *,
        context: ProxyContext,
        scope: Scope,
        receive: Receive,
) -> aiohttp.ClientResponse:
    request = Request(scope, receive)
    should_stream_incoming = determine_incoming_streaming(request)
    async with context.semaphore:
        data: Union[None, AsyncGenerator[bytes, None], bytes] = None
        if request.method not in ("GET", "HEAD"):
            if should_stream_incoming:
                data = request.stream()
            else:
                data = await request.body()

        kwargs = context.config.get_upstream_http_options(
            scope=scope, client_request=request, data=data
        )

        return await context.session.request(**kwargs)


async def convert_proxy_response_to_user_response(
        *,
        context: ProxyContext,
        scope: Scope,
        proxy_response: aiohttp.ClientResponse,
) -> Response:
    headers_to_client = context.config.process_upstream_headers(
        scope=scope, proxy_response=proxy_response
    )
    status_to_client = proxy_response.status
    if determine_outgoing_streaming(proxy_response):
        return StreamingResponse(
            content=read_stream_in_chunks(proxy_response.content),
            status_code=status_to_client,
            headers=headers_to_client,  # type: ignore
        )
    new_headers = headers_to_client
    response_content = await proxy_response.read()

    # Forked code
    # only rewrite for databricks proxy
    if is_from_databricks_proxy(scope) is True:
        if response_content is not None and len(response_content) > 0 and context.config.modify_content is not None:
            for path_pattern, modify_func in context.config.modify_content.items():
                # if path is .js it should be of type text/javascript;charset=utf-8
                if fnmatch.fnmatch(scope["path"], path_pattern):
                    response_content = modify_func(response_content)
                    new_headers.popall("Content-Length", None)
                    new_headers["Content-Length"] = str(len(response_content))
                    # TODO: this may cause bugs :\ if we need multiple passes
                    # TODO: in future maybe we have priority or flag
                    break

    return Response(
        content=response_content,
        status_code=status_to_client,
        headers=new_headers,  # type: ignore
    )

async def get_user_response(*,
        context: ProxyContext,
        scope: Scope,
        receive: Receive):
    proxy_response = await get_proxy_response(
        context=context, scope=scope, receive=receive
    )
    user_response = await convert_proxy_response_to_user_response(
        context=context, scope=scope, proxy_response=proxy_response
    )
    return user_response

async def proxy_http(
        *,
        context: ProxyContext,
        scope: Scope,
        receive: Receive,
        send: Send,
) -> None:
    root_path = scope["root_path"]
    if is_from_databricks_proxy(scope) is True:
        if scope["path"].startswith(root_path):
            scope["path"] = scope["path"].replace(root_path, "")

    # try to get response twice before sending 502
    try:
        user_response = await get_user_response(
            context=context, scope=scope, receive=receive
        )
    except aiohttp.client_exceptions.ClientConnectorError as cce:
        print(f"Failed to connect to server; retrying in 1 second: {str(cce)}")
        await asyncio.sleep(1)
        try:
            user_response = await get_user_response(
                context=context, scope=scope, receive=receive
            )
        except aiohttp.client_exceptions.ClientConnectorError as cce:
            print(f"Failed to connect to server the second time for same connection; retrying in 1 second: {str(cce)}")
            user_response = Response(
                status_code=502,
                content="Unable to connect to app waiting for app server to respond. "
                        "Refresh a few times otherwise restart.",
            )

    return await user_response(scope, receive, send)



from dbtunnel.tunnels import DbTunnel


class DashAppTunnel(DbTunnel):

    def __init__(self, dash_app, port: int = 8080):
        super().__init__(port, flavor="dash")
        self._dash_app = dash_app

    def _imports(self):
        try:
            from fastapi import FastAPI
            import uvicorn
            import dash
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have fastapi, nest_asyncio, dash and uvicorn installed;"
                  "pip install fastapi nest_asyncio dash uvicorn")
            raise e

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        from fastapi import FastAPI
        import uvicorn
        from fastapi.middleware.wsgi import WSGIMiddleware
        app = FastAPI(root_path=self._proxy_settings.url_base_path.rstrip("/"))
        # DASH HACK
        del self._dash_app.config._read_only['requests_pathname_prefix']
        del self._dash_app.config._read_only['routes_pathname_prefix']
        self._dash_app.config.update({
            'routes_pathname_prefix': self._proxy_settings.url_base_path,
            'requests_pathname_prefix': self._proxy_settings.url_base_path
        })
        app.mount("/", WSGIMiddleware(self._dash_app.server))
        import nest_asyncio
        nest_asyncio.apply()

        # uvicorn.run(app, host="0.0.0.0", port=self._port)
        # Start the server
        async def start():
            config = uvicorn.Config(
                app,
                host="0.0.0.0",
                port=self._port,
            )
            server = uvicorn.Server(config)
            await server.serve()

        # Run the asyncio event loop instead of uvloop to enable re entrance
        import asyncio
        self._log.info(f"Use this link: \n{self._proxy_settings.proxy_url}")
        asyncio.run(start())

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="dbtunnel",
    author="Sri Tikkireddy",
    author_email="sri.tikkireddy@databricks.com",
    description="Run app and get cluster proxy url for it in databricks clusters",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={
        'dbtunnel': ['**/*.html'],
    },
    url="https://github.com/stikkireddy/dbtunnel",
    packages=find_packages(),
    install_requires=["nest_asyncio", "databricks-sdk>=0.18.0"],
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    # TODO: start refactoring nest_asyncio only when async.run is needed
    extras_require={
        "fastapi": [
            # Specify dependencies for building documentation here
            "fastapi",
            "uvicorn",
        ],
        "uvicorn": [
            # Specify dependencies for building documentation here
            "uvicorn",
        ],
        "streamlit": [
            # Specify dependencies for building documentation here
            "streamlit",
            "pyarrow>=11",
        ],
        "gradio": [
            # Specify dependencies for building documentation here
            "gradio==3.50.2",
        ],
        "nicegui": [
            # Specify dependencies for building documentation here
            "nicegui",
        ],
        "bokeh": [
            "bokeh",
        ],
        "flask": [
            "flask",
            "fastapi",
            "uvicorn",  # no websockets
        ],
        "dash": [
            "dash",
            "fastapi",
            "uvicorn",  # no websockets
        ],
        "sql": [
            "databricks-sql-connector"
        ],
        "solara": [
            "solara",
        ],
        "ngrok": [
            # going back to pyngrok its more stable and easy to delete sessions
            "pyngrok",
            "ngrok-api",
            "requests",
        ],
        "chainlit": [
            "chainlit",
            "uvicorn",
        ],
        "asgiproxy": [
            "aiohttp",
            "starlette",
            "uvicorn",
            "websockets",
            "python-multipart",  # we are using this for auth check via form uploads
            "cachetools"
        ],
        "shiny": [
            "shiny",
        ],
        "dev": [
            "mkdocs-material",
            "mkdocs-jupyter",
        ],
        "cli": [
            "click",
        ],
        "arize-phoenix": [
            "arize-phoenix"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'dbtunnel = dbtunnel.cli.cli:cli',
        ],
    },
    python_requires=">=3.10",
)



# pip install langchain-community tiktoken langchain mlflow-skinny chromadb rank_bm25 simsimd

import asyncio
from typing import TypeVar, Protocol, AsyncIterator
import chainlit as cl
from databricks_genai_inference.api.chat_completion import ChatCompletion
from langchain.retrievers import MergerRetriever, ContextualCompressionRetriever
from langchain.retrievers.document_compressors import DocumentCompressorPipeline
from langchain.text_splitter import CharacterTextSplitter
from langchain_community.document_loaders import GutenbergLoader
from langchain_community.document_transformers import EmbeddingsRedundantFilter, LongContextReorder
from langchain_community.embeddings import DatabricksEmbeddings
from langchain_community.retrievers import BM25Retriever
from langchain_community.vectorstores.chroma import Chroma
from langchain_core.documents import Document
from langchain_core.prompts import ChatPromptTemplate, HumanMessagePromptTemplate, PromptTemplate
from langchain_core.runnables import RunnablePassthrough


class HasMessage(Protocol):
    message: str


T = TypeVar('T', bound=HasMessage)


class AsyncGeneratorWrapper(AsyncIterator[T]):
    def __init__(self, gen):
        self.gen = gen

    def __aiter__(self) -> 'AsyncGeneratorWrapper[T]':
        return self

    async def __anext__(self) -> T:
        try:
            # Use asyncio to yield control and create asynchronous behavior
            await asyncio.sleep(0)
            return next(self.gen)
        except StopIteration:
            raise StopAsyncIteration


def format_docs(docs):
    return "\n\n".join(
        f"{doc.page_content}"
        for doc in docs)


v0_template = """
[INST] You are going to answer questions about the book of Frankenstein; Or, The Modern Prometheus by Mary Wollstonecraft Shelley
You will not answer anything else.

Here's some context which might or might not help you answer: {context}

Based on this context, answer this question: {question}
[/INST]
"""

prompt = ChatPromptTemplate(input_variables=['context', 'question'],
                            messages=[
                                HumanMessagePromptTemplate(
                                    prompt=PromptTemplate(
                                        input_variables=['context', 'question'],
                                        template=v0_template))])

loader = GutenbergLoader("https://www.gutenberg.org/cache/epub/84/pg84.txt")
docs = loader.load()
text_splitter = CharacterTextSplitter.from_tiktoken_encoder(
    chunk_size=512, chunk_overlap=100
)
docs = [Document(page_content=text) for text in text_splitter.split_text(docs[0].page_content)]
embeddings = DatabricksEmbeddings(endpoint="databricks-bge-large-en")
chroma = Chroma.from_documents(docs, embeddings)
chroma_retriever = chroma.as_retriever(search_type="mmr", search_kwargs={"k": 5, "include_metadata": True})
bm25_retriever = BM25Retriever.from_documents(docs, search_type="mmr", search_kwargs={"k": 5, "include_metadata": True})
lotr = MergerRetriever(retrievers=[chroma_retriever, bm25_retriever])
#
filter_ = EmbeddingsRedundantFilter(embeddings=embeddings)
reordering = LongContextReorder()
pipeline = DocumentCompressorPipeline(transformers=[filter_, reordering])
compression_retriever = ContextualCompressionRetriever(
    base_compressor=pipeline, base_retriever=lotr
)

rag_prompt = (
        {"context": compression_retriever | format_docs, "question": RunnablePassthrough()}
        | prompt
)


def run_chat_completion(msgs) -> AsyncGeneratorWrapper[HasMessage]:
    resp = ChatCompletion.create(model="mixtral-8x7b-instruct",
                                 messages=[{"role": "system", "content": "You are a helpful assistant."},
                                           *msgs],
                                 temperature=0.1,
                                 stream=True, )
    return AsyncGeneratorWrapper(resp)


@cl.step
async def retriever(content: str):
    loop = asyncio.get_event_loop()
    await cl.sleep(0)
    processed_prompt = await loop.run_in_executor(None, rag_prompt.invoke, content)
    return [{"content": msg.content, "role": "user"} for msg in processed_prompt.to_messages()]


@cl.on_chat_start
async def on_start():
    msg = cl.Message(content="Please ask any questions about the book of Frankenstein; Or, The Modern Prometheus by Mary Wollstonecraft Shelley. Example: Who is Frankenstein's monster?")

    await msg.send()


@cl.on_message
async def main(message: cl.Message):
    # Your custom logic goes here...
    response = cl.Message(content="")
    loop = asyncio.get_event_loop()
    await response.send()  # causes loading indicator
    msgs = await retriever(message.content)
    token_stream = await loop.run_in_executor(None, run_chat_completion, msgs)
    async for token_chunk in token_stream:
        chunk: HasMessage = token_chunk  # noqa token_chunk is a ChatCompletionChunkObject not Future
        await response.stream_token(chunk.message)

    # Send a response back to the user
    await response.update()



import atexit
import datetime
import logging
import os
import queue
import shutil
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from fnmatch import fnmatch
from functools import cached_property
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import List, Any, Optional, Literal

try:
    from databricks.sdk import WorkspaceClient
except ImportError:
    print("databricks-sdk not installed. Please install databricks-sdk to use this feature")
    raise


@contextmanager
def process_file(input_path):
    temp_dir = tempfile.mkdtemp()
    try:
        # Create a temporary directory

        # Build the destination path in the temporary directory
        temp_file_path = os.path.join(temp_dir, os.path.basename(input_path))

        # Copy the file to the temporary directory
        shutil.copy(input_path, temp_file_path)

        # Yield the temporary file path to the caller
        yield temp_file_path
    finally:
        # Cleanup: Remove the temporary directory and its contents
        shutil.rmtree(temp_dir, ignore_errors=True)


def ensure_python_path(env):
    import sys
    from pathlib import Path
    for python_version_dir in (Path(sys.executable).parent.parent / "lib").iterdir():
        site_packages = str(python_version_dir / "site-packages")
        py_path = env.get("PYTHONPATH", "")
        if site_packages not in py_path.split(":"):
            env["PYTHONPATH"] = f"{py_path}:{site_packages}"


def execute(cmd: List[str], env, cwd=None, ensure_python_site_packages=True, shell=False, trim_new_line=True):
    if ensure_python_site_packages:
        ensure_python_path(env)
    import subprocess
    if shell is True:
        cmd = " ".join(cmd)
    popen = subprocess.Popen(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True,
                             shell=shell,
                             env=env,
                             cwd=cwd,
                             bufsize=1)
    if popen.stdout is not None:
        for stdout_line in iter(popen.stdout.readline, ""):
            if trim_new_line:
                stdout_line = stdout_line.strip()
            yield stdout_line

    # if popen.stderr is not None:
    #     for stderr_line in iter(popen.stderr.readline, ""):  # Iterate over stderr
    #         yield stderr_line

    popen.stdout.close()
    # popen.stderr.close()  # Close stderr
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def pkill(process_name):
    try:
        subprocess.run(["pkill", process_name])
        print(f"Processes with name '{process_name}' killed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error killing processes: {e}")

# from langchain: https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/llms/databricks.py#L86
def get_repl_context() -> Any:
    """Gets the notebook REPL context if running inside a Databricks notebook.
    Returns None otherwise.
    """
    try:
        from dbruntime.databricks_repl_context import get_context

        return get_context()
    except ImportError:
        raise ImportError(
            "Cannot access dbruntime, not running inside a Databricks notebook."
        )


def get_workspace_host_via_spark_config() -> Optional[str]:
    try:
        from pyspark.sql import SparkSession

        # databricks notebook and jobs will already have spark sessions
        spark = SparkSession.getActiveSession()
        return spark.conf.get("spark.databricks.workspaceUrl")
    except Exception:
        print("Not running inside a Databricks notebook or job, unable to get workspace url from spark session.")
        return None


class DatabricksContext:

    def __init__(self):
        self._repl_ctx = get_repl_context()

    @cached_property
    def host(self) -> str:
        if hasattr(self._repl_ctx, "browserHostName"):
            return self._repl_ctx.browserHostName
        spark_conf_ws_url = get_workspace_host_via_spark_config()
        if spark_conf_ws_url is not None:
            return spark_conf_ws_url
        raise ValueError("Unable to get workspace host from REPL context or spark config")

    @cached_property
    def token(self) -> str:
        if hasattr(self._repl_ctx, "apiToken"):
            return self._repl_ctx.apiToken
        raise ValueError("Unable to get token from REPL context")

    @cached_property
    def current_user_name(self) -> str:
        return WorkspaceClient(host=self.host,
                               token=self.token).current_user.me().user_name

    @cached_property
    def current_username_alphanumeric(self) -> str:
        sanitized = self.current_user_name.split("@")[0].replace(".", "-")
        return "".join(c for c in sanitized if c.isalnum() or c == "-")


@dataclass
class WarehouseDetails:
    hostname: str
    http_path: str
    name: str
    serverless: bool


class ComputeUtils:

    def __init__(self, dbx_ctx: DatabricksContext):
        self._ctx = dbx_ctx
        self._client = WorkspaceClient(host=self._ctx.host, token=self._ctx.token)

    def get_warehouse(self, name_glob: str, ignore_case: bool = True, serverless_only: bool = True):
        for warehouse in self._client.warehouses.list():
            if serverless_only is True and warehouse.enable_serverless_compute is False:
                continue  # Skip warehouses that do not have serverless compute enabled

            if ignore_case is True:
                name_match = fnmatch(warehouse.name.lower(), name_glob.lower())
            else:
                name_match = fnmatch(warehouse.name, name_glob)

            if name_match is False:
                continue  # Skip warehouses that do not match the name_glob

            hostname = warehouse.odbc_params.hostname
            http_path = warehouse.odbc_params.path
            return WarehouseDetails(hostname,
                                    http_path,
                                    warehouse.name,
                                    warehouse.enable_serverless_compute)


class ArchivingTimedRotatingFileHandler(TimedRotatingFileHandler):
    """
    Rotate the files in the cluster native FS and during the rotate time period copy the file to
    a volume so there are no issues with file system shenanigans with FUSE implementation
    """

    def __init__(self, archive_path: Path, filename, when='h', interval=1, backupCount=0,
                 encoding=None, delay=False, utc=False, atTime=None,
                 errors=None):
        super().__init__(filename, when, interval, backupCount, encoding, delay, utc, atTime, errors)
        self._archive_path = archive_path
        if not self._archive_path.exists():
            self._archive_path.mkdir(parents=True)

    def rotate(self, source, dest):
        super().rotate(source, dest)
        # copy dest file to archive path
        self.archive_log_file(dest)

    def archive_log_file(self, log_file):
        try:
            shutil.copy(log_file, str(self._archive_path))
        except Exception as e:
            print(f"Unable to archive log file: {e}")


def get_logger(
        *,
        app_name: str = "dbtunnel",
        cluster_logging_file_path: Optional[Path] = None,
        logging_archive_folder: Optional[Path] = None,
        rotate_when: Literal["S", "M", "H", "D", "midnight"] = "H",
        rotate_interval: int = 1,
        backup_count: int = 3,
        at_time: Optional[datetime.time] = None,
        format_str: str = "[%(asctime)s] [%(levelname)s] {%(module)s.py:%(funcName)s:%(lineno)d} - %(message)s",
        datefmt_str: str = "%Y-%m-%dT%H:%M:%S%z"
):
    # driver instead of workspace path to not deal with WSFS
    home_dir = os.path.expanduser('~')
    cluster_logging_file_path = cluster_logging_file_path or Path(f"{home_dir}/logs/{app_name}/{app_name}.log")
    if not cluster_logging_file_path.parent.exists():
        cluster_logging_file_path.parent.mkdir(parents=True)

    logger = logging.getLogger(app_name)

    # Create a queue
    log_queue = queue.Queue(maxsize=100)

    # Create a formatter for 'simple' and 'detailed' formats
    detailed_formatter = logging.Formatter(
        format_str,
        datefmt=datefmt_str
    )

    # Create a StreamHandler for 'stderr' with 'simple' formatter
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.INFO)
    stdout_handler.setFormatter(detailed_formatter)

    # Create a RotatingFileHandler for 'file' with 'detailed' formatter
    time_rotate_cfg = {
        "when": rotate_when,
        "interval": rotate_interval,
        "backupCount": backup_count,
        "atTime": at_time,
    }
    if logging_archive_folder is not None and logging_archive_folder.exists():
        file_handler = ArchivingTimedRotatingFileHandler(
            archive_path=logging_archive_folder,
            filename=cluster_logging_file_path,
            **time_rotate_cfg
        )
    else:
        file_handler = logging.handlers.TimedRotatingFileHandler(
            filename=cluster_logging_file_path,
            **time_rotate_cfg
        )
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(detailed_formatter)

    # Create a QueueListener with the created queue and the added handlers
    queue_handler = logging.handlers.QueueHandler(log_queue)

    queue_listener = logging.handlers.QueueListener(
        log_queue, stdout_handler, file_handler, respect_handler_level=True
    )

    # incase this is reinitialized remove all handlers
    for handler in logger.handlers:
        logger.removeHandler(handler)

    # Add the QueueListener handler to the root logger
    logger.addHandler(queue_handler)

    # Set the log level for the root logger
    logger.setLevel(logging.DEBUG)

    # disable py4j logger
    logging.getLogger("py4j").setLevel(logging.ERROR)

    # Start the QueueListener
    queue_listener.start()

    # Register a function to stop the QueueListener on program exit
    atexit.register(queue_listener.stop)

    return logger


try:
    ctx = DatabricksContext()
    compute_utils = ComputeUtils(ctx)
except Exception as e:
    logging.info("Unable to establish context, you are most likely running outside of databricks")
    ctx = None
    compute_utils = None



from dbtunnel.tunnels import DbTunnel
from dbtunnel.utils import execute, process_file


class SolaraAppTunnel(DbTunnel):

    def __init__(self, script_path: str, port: int):
        super().__init__(port, "solara")
        self._script_path = script_path

    def _imports(self):
        try:
            import solara
        except ImportError as e:
            self._log.info("ImportError: Make sure you have solara installed. \n"
                  "pip install solara")
            raise e

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        self._log.info(f"Use this link: \n{self._proxy_settings.proxy_url}")
        with process_file(self._script_path) as file_path:
            self.run_solara(file_path, self._port)

    def run_solara(self, path, port):
        import os
        import subprocess
        my_env = os.environ.copy()
        subprocess.run(f"kill -9 $(lsof -t -i:{port})", capture_output=True, shell=True)
        # static assets otherwise get served by root and the root path is not allowed!
        self._log.info(f"Deploying {self._flavor} app at path: {path} on port: {port}")
        server_path_prefix = self._proxy_settings.url_base_path.rstrip('/')
        cmd = [
               "solara",
               "run",
               "--host",
               "0.0.0.0",
               "--port",
               str(port),
               "--production",
               "--root-path",
               server_path_prefix,
               path,
            ]
        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path)



from dbtunnel.tunnels import DbTunnel, DbTunnelProxy
from dbtunnel.utils import execute
from dbtunnel.vendor.asgiproxy.frameworks import Frameworks

class ChainlitAppTunnel(DbTunnel):

    def _imports(self):
        try:
            import chainlit
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have chainlit installed. \n"
                  "pip install chainlit nest_asyncio")
            raise e

    def _run(self):
        import os

        chainlit_service_port_no_share = 9090
        proxy_service = None
        if self._share is False:

            url_base_path = self._proxy_settings.url_base_path
            port = self._port

            # nest uvicorn to the ipynotebook asyncio eventloop so restarting kernel kills server
            proxy_service = DbTunnelProxy(
                proxy_port=port,
                service_port=chainlit_service_port_no_share,
                url_base_path=url_base_path,
                framework=Frameworks.CHAINLIT,
                token_auth=self._basic_tunnel_auth["token_auth"],
                token_auth_workspace_url=self._basic_tunnel_auth["token_auth_workspace_url"],
                cwd=self._cwd
            )

            proxy_service.start()

        self._log.info(f"Use this link to access the Chainlit UI in Databricks: \n{self._proxy_settings.proxy_url}")

        self._log.info("Starting chainlit...")

        my_env = os.environ.copy()

        if self._share is False:
            cmd = ["chainlit", "run", self._chainlit_script_path, "-h", "--host", "0.0.0.0", "--port",
                   f"{chainlit_service_port_no_share}"]
        else:
            cmd = ["chainlit", "run", self._chainlit_script_path, "-h", "--host", "0.0.0.0", "--port", f"{self._port}"]

        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env, cwd=self._cwd):
            self._log.info(path)

        if proxy_service is not None:
            proxy_service.wait()

    def __init__(self, chainlit_script_path: str, cwd: str = None, port: int = 8000):
        super().__init__(port, "chainlit")
        self._chainlit_script_path = chainlit_script_path
        self._cwd = cwd

    def _display_url(self):
        return None



from dbtunnel.tunnels import DbTunnel

class ShinyPythonAppTunnel(DbTunnel):

    def __init__(self, shiny_app, port: int = 8080):
        super().__init__(port, flavor="shiny-python")
        self._shiny_app = shiny_app

    def _imports(self):
        try:
            import shiny
            import uvicorn
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have shiny, nest_asyncio and uvicorn installed;"
                  "pip install fastapi nest_asyncio uvicorn")
            raise e

    def _run(self):
        self.display()
        self._log.info("Starting server...")
        import uvicorn
        import nest_asyncio
        nest_asyncio.apply()

        async def start():
            config = uvicorn.Config(
                self._shiny_app,
                host="0.0.0.0",
                port=self._port,
            )
            server = uvicorn.Server(config)
            await server.serve()

        # Run the asyncio event loop instead of uvloop to enable re entrance
        import asyncio
        self._log.info(f"Use this link: \n{self._proxy_settings.get_proxy_url(ensure_ends_with_slash=True)}")
        asyncio.run(start())

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'



import os
import subprocess

from dbtunnel.tunnels import DbTunnel, get_current_username
from dbtunnel.utils import execute


class CodeServerTunnel(DbTunnel):

    def __init__(self,
                 directory_path: str = None,
                 repo_name: str = None,
                 port: int = 9988,
                 extension_ids: list[str] = None):
        # Check if either directory_path or repo_name is provided
        if directory_path is None and repo_name is None:
            raise ValueError("Either directory_path or repo_name must be provided.")

        if repo_name:
            repo_path = f"/Workspace/Repos/{get_current_username()}/{repo_name}"
            if not os.path.exists(repo_path):
                raise ValueError(f"The provided repo_name '{repo_name}' in path: {repo_path} does not exist.")
            self._dir_path = repo_path

        # Check if directory_path is provided and exists
        if directory_path:
            if not os.path.exists(directory_path):
                raise ValueError(f"The provided directory_path '{directory_path}' does not exist.")
            self._dir_path = directory_path

        self._extension_ids = extension_ids or []

        # Continue with the initialization
        super().__init__(port, "code-server")

    def _imports(self):
        return None

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _install_databricks_cli(self):
        self._log.info("Attempting to install databricks cli")
        command = "curl -fsSL https://raw.githubusercontent.com/databricks/setup-cli/main/install.sh | sh"
        env_copy = os.environ.copy()
        already_installed = False
        try:
            for stmt in execute([command], shell=True, env=env_copy):
                if "already exists" in stmt:
                    already_installed = True
                self._log.info(stmt)
        except subprocess.CalledProcessError as e:
            if already_installed is False:
                raise e
        self._log.info("Finished installing databricks cli")

    def _install_extension(self, env, extension_id: str):
        for stmt in execute(["code-server", "--install-extension", extension_id], shell=True, env=env):
            self._log.info(stmt)
        self._log.info(f"Finished Installed extension: {extension_id}")

    def _install_extensions(self, env, extensions: list[str]):
        for extension in extensions:
            self._install_extension(env, extension)

    def _run(self):
        import subprocess

        self._log.info(f"Use this link: \n{self._proxy_settings.proxy_url}?folder={self._dir_path}")
        self._log.info("It may take a 15-30 seconds for the code server to start up.")

        self._log.info("Installing code server")
        url = "https://code-server.dev/install.sh"
        # Equivalent Python subprocess command with piping
        subprocess.run(f'curl -fsSL {url} | sh', check=True, shell=True)
        self._log.info("Installed code server")

        self._install_databricks_cli()

        import os
        import subprocess
        my_env = os.environ.copy()
        my_env["VSCODE_PROXY_URI"] = self._proxy_settings.url_base_path + "wss"
        subprocess.run(f"kill -9 $(lsof -t -i:{self._port})", capture_output=True, shell=True)

        self._log.info(f"Installing default plugins!")
        default_plugins = [
            "ms-python.python",
            "ms-toolsai.jupyter",
            "ms-toolsai.jupyter-renderers",
            "ms-toolsai.jupyter-keymap",
            "ms-toolsai.vscode-jupyter-cell-tags",
            "databricks.databricks",
            "databricks.sqltools-databricks-driver",
            "rangav.vscode-thunder-client",
        ]
        self._install_extensions(my_env, list(set(default_plugins + self._extension_ids)))

        # "VSCODE_PROXY_URI=“./driver-proxy/o/<id>/1201-175053-rt06lneb/8080/wss” code-server --bind-addr 0.0.0.0:8080  --auth none"
        self._log.info(f"Deploying code server on port: {self._port}")
        cmd = ["code-server",
               "--bind-addr",
               f"0.0.0.0:{self._port}",
               "--auth",
               "none"]
        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path)


# Databricks notebook source
# MAGIC %pip install dbtunnel

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

from dbtunnel import dbtunnel
# alternatively you can do directory_path= for full directory
# dbtunnel.code_server(directory_path="/Workspace/...").run()
# no need for auth you should be using single user clusters!
dbtunnel.code_server(repo_name="dbtunnel").run()

# COMMAND ----------





# Databricks notebook source
# MAGIC %pip install dbtunnel[asgiproxy,gradio]

# COMMAND ----------

# MAGIC %pip install -U gradio==4.16.0

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/gradio_example.py"
if os.path.exists(script_path) is False:
  raise Exception(f"File doesnt exist: {script_path}") 

# COMMAND ----------

from dbtunnel import dbtunnel

dbtunnel.gradio(path=script_path).run()

# COMMAND ----------





import os
import subprocess
import sys

from dbtunnel.tunnels import DbTunnel, DbTunnelProxy
from dbtunnel.utils import execute
from dbtunnel.vendor.asgiproxy.frameworks import Frameworks


class ArizePhoenixUITunnel(DbTunnel):

    def __init__(self, port: int):
        super().__init__(port, "arize-phoenix")


    def _imports(self):
        try:
            import phoenix.server
            import nest_asyncio
        except ImportError as e:
            self._log.info("ImportError: Make sure you have chainlit installed. \n"
                           "pip install arize-phoenix nest_asyncio")
            raise e

    def _display_url(self):
        # must end with a "/" for it to not redirect
        return f'<a href="{self._proxy_settings.proxy_url}">Click to go to {self._flavor} App!</a>'

    def _run(self):
        self.display()
        self._log.info("Starting server...")

        phoenix_service_port_no_share = 9099
        proxy_service = None
        if self._share is False:
            url_base_path = self._proxy_settings.url_base_path
            port = self._port

            proxy_service = DbTunnelProxy(
                proxy_port=port,
                service_port=phoenix_service_port_no_share,
                url_base_path=url_base_path,
                framework=Frameworks.ARIZE_PHOENIX,
                token_auth=self._basic_tunnel_auth["token_auth"],
                token_auth_workspace_url=self._basic_tunnel_auth["token_auth_workspace_url"],
            )

            proxy_service.start()

        my_env = os.environ.copy()
        subprocess.run(f"kill -9 $(lsof -t -i:{self._port})", capture_output=True, shell=True)

        if self.shared is False:
            self._log.info(f"Deploying stable diffusion web ui app at path: \n{self._proxy_settings.proxy_url}")

        if self._share is False:
            cmd = [sys.executable, "-m", "phoenix.server.main", "--port", f"{phoenix_service_port_no_share}",
                   "serve"]
        else:
            cmd = [sys.executable, "-m", "phoenix.server.main", "--port", f"{self._port}", "serve"]

        self._log.info(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env):
            self._log.info(path)

        if proxy_service is not None:
            proxy_service.wait()



# Databricks notebook source
# MAGIC %pip install dbtunnel[asgiproxy,chainlit,ngrok]

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/chainlit_example.py"

# COMMAND ----------

from dbtunnel import dbtunnel

dbtunnel.chainlit(script_path).share_to_internet_via_ngrok(
    ngrok_api_token="<ngrok api token>",
    ngrok_tunnel_auth_token="<ngrok tunnel auth token>"
).run()

# COMMAND ----------


